@@ -3,12 +3,12 @@
  * rtscan.c
  *	  routines to manage scans on index relations
  *
- * Portions Copyright (c) 1996-2000, PostgreSQL, Inc
+ * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /usr/local/cvsroot/pgsql/src/backend/access/rtree/rtscan.c,v 1.32 2000/04/12 17:14:51 momjian Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/access/rtree/rtscan.c,v 1.39 2001/10/25 05:49:21 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -51,49 +51,37 @@ typedef RTScanListData *RTScanList;
 /* pointer to list of local scans on rtrees */
 static RTScanList RTScans = (RTScanList) NULL;
 
-IndexScanDesc
-rtbeginscan(Relation r,
-			bool fromEnd,
-			uint16 nkeys,
-			ScanKey key)
+Datum
+rtbeginscan(PG_FUNCTION_ARGS)
 {
+	Relation	r = (Relation) PG_GETARG_POINTER(0);
+	bool		fromEnd = PG_GETARG_BOOL(1);
+	uint16		nkeys = PG_GETARG_UINT16(2);
+	ScanKey		key = (ScanKey) PG_GETARG_POINTER(3);
 	IndexScanDesc s;
 
-	/*
-	 * Let index_beginscan does its work...
-	 *
-	 * RelationSetLockForRead(r);
-	 */
-
 	s = RelationGetIndexScan(r, fromEnd, nkeys, key);
+
 	rtregscan(s);
 
-	return s;
+	PG_RETURN_POINTER(s);
 }
 
-void
-rtrescan(IndexScanDesc s, bool fromEnd, ScanKey key)
+Datum
+rtrescan(PG_FUNCTION_ARGS)
 {
+	IndexScanDesc s = (IndexScanDesc) PG_GETARG_POINTER(0);
+	bool		fromEnd = PG_GETARG_BOOL(1);
+	ScanKey		key = (ScanKey) PG_GETARG_POINTER(2);
 	RTreeScanOpaque p;
 	RegProcedure internal_proc;
 	int			i;
 
-	if (!IndexScanIsValid(s))
-	{
-		elog(ERROR, "rtrescan: invalid scan.");
-		return;
-	}
-
 	/*
 	 * Clear all the pointers.
 	 */
-
-	ItemPointerSetInvalid(&s->previousItemData);
 	ItemPointerSetInvalid(&s->currentItemData);
-	ItemPointerSetInvalid(&s->nextItemData);
-	ItemPointerSetInvalid(&s->previousMarkData);
 	ItemPointerSetInvalid(&s->currentMarkData);
-	ItemPointerSetInvalid(&s->nextMarkData);
 
 	/*
 	 * Set flags.
@@ -157,11 +145,14 @@ rtrescan(IndexScanDesc s, bool fromEnd, ScanKey key)
 			}
 		}
 	}
+
+	PG_RETURN_VOID();
 }
 
-void
-rtmarkpos(IndexScanDesc s)
+Datum
+rtmarkpos(PG_FUNCTION_ARGS)
 {
+	IndexScanDesc s = (IndexScanDesc) PG_GETARG_POINTER(0);
 	RTreeScanOpaque p;
 	RTSTACK    *o,
 			   *n,
@@ -190,11 +181,14 @@ rtmarkpos(IndexScanDesc s)
 
 	freestack(p->s_markstk);
 	p->s_markstk = o;
+
+	PG_RETURN_VOID();
 }
 
-void
-rtrestrpos(IndexScanDesc s)
+Datum
+rtrestrpos(PG_FUNCTION_ARGS)
 {
+	IndexScanDesc s = (IndexScanDesc) PG_GETARG_POINTER(0);
 	RTreeScanOpaque p;
 	RTSTACK    *o,
 			   *n,
@@ -223,11 +217,14 @@ rtrestrpos(IndexScanDesc s)
 
 	freestack(p->s_stack);
 	p->s_stack = o;
+
+	PG_RETURN_VOID();
 }
 
-void
-rtendscan(IndexScanDesc s)
+Datum
+rtendscan(PG_FUNCTION_ARGS)
 {
+	IndexScanDesc s = (IndexScanDesc) PG_GETARG_POINTER(0);
 	RTreeScanOpaque p;
 
 	p = (RTreeScanOpaque) s->opaque;
@@ -241,6 +238,8 @@ rtendscan(IndexScanDesc s)
 
 	rtdropscan(s);
 	/* XXX don't unset read lock -- two-phase locking */
+
+	PG_RETURN_VOID();
 }
 
 static void
@@ -278,6 +277,27 @@ rtdropscan(IndexScanDesc s)
 	pfree(l);
 }
 
+/*
+ * AtEOXact_rtree() --- clean up rtree subsystem at xact abort or commit.
+ *
+ * This is here because it needs to touch this module's static var RTScans.
+ */
+void
+AtEOXact_rtree(void)
+{
+	/*
+	 * Note: these actions should only be necessary during xact abort; but
+	 * they can't hurt during a commit.
+	 */
+
+	/*
+	 * Reset the active-scans list to empty. We do not need to free the
+	 * list elements, because they're all palloc()'d, so they'll go away
+	 * at end of transaction anyway.
+	 */
+	RTScans = NULL;
+}
+
 void
 rtadjscans(Relation r, int op, BlockNumber blkno, OffsetNumber offnum)
 {
@@ -360,7 +380,6 @@ adjustiptr(IndexScanDesc s,
 						}
 						else
 						{
-
 							/*
 							 * remember that we're before the current
 							 * tuple
