@@ -13,7 +13,7 @@
  * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
- * $Header: /cvsroot/pgsql-server/src/backend/access/transam/clog.c,v 1.7 2001/10/28 06:25:42 momjian Exp $
+ * $Header: /cvsroot/pgsql-server/src/backend/access/transam/clog.c,v 1.7.2.1 2002/09/30 19:55:08 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -810,8 +810,8 @@ TruncateCLOG(TransactionId oldestXact)
 	if (!ScanCLOGDirectory(cutoffPage, false))
 		return;					/* nothing to remove */
 
-	/* Perform a CHECKPOINT */
-	CreateCheckPoint(false);
+	/* Perform a forced CHECKPOINT */
+	CreateCheckPoint(false, true);
 
 	/*
 	 * Scan CLOG shared memory and remove any pages preceding the cutoff
