@@ -26,10 +26,10 @@
  * the backend's "backend/libpq" is quite separate from "interfaces/libpq".
  * All that remains is similarities of names to trap the unwary...
  *
- * Portions Copyright (c) 1996-2000, PostgreSQL, Inc
+ * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
- *	$Id: pqcomm.c,v 1.92 2000/05/26 01:26:19 tgl Exp $
+ *	$Id: pqcomm.c,v 1.126 2001/12/04 20:57:22 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -41,13 +41,13 @@
  *		StreamServerPort	- Open postmaster's server port
  *		StreamConnection	- Create new connection with client
  *		StreamClose			- Close a client/backend connection
- *		pq_getport		- return the PGPORT setting
  *		pq_init			- initialize libpq at backend startup
  *		pq_close		- shutdown libpq at backend exit
  *
  * low-level I/O:
  *		pq_getbytes		- get a known number of bytes from connection
  *		pq_getstring	- get a null terminated string from connection
+ *		pq_getbyte		- get next byte from connection
  *		pq_peekbyte		- peek at next byte from connection
  *		pq_putbytes		- send bytes to connection (not flushed until pq_flush)
  *		pq_flush		- flush pending output
@@ -59,31 +59,37 @@
  *
  *------------------------
  */
+#include "postgres.h"
+
 #include <signal.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <grp.h>
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/socket.h>
 #include <netdb.h>
 #include <netinet/in.h>
+#ifdef HAVE_NETINET_TCP_H
 #include <netinet/tcp.h>
+#endif
 #include <arpa/inet.h>
 #include <sys/file.h>
 
-#include "postgres.h"
-
 #include "libpq/libpq.h"
-#include "utils/trace.h"		/* needed for HAVE_FCNTL_SETLK */
 #include "miscadmin.h"
 
 
-#ifndef SOMAXCONN
-#define SOMAXCONN 5				/* from Linux listen(2) man page */
-#endif	 /* SOMAXCONN */
+static void pq_close(void);
+
+
+/*
+ * Configuration options
+ */
+int			Unix_socket_permissions;
+char	   *Unix_socket_group;
 
-extern FILE *debug_port;		/* in util.c */
 
 /*
  * Buffers for low-level I/O
@@ -115,23 +121,9 @@ pq_init(void)
 {
 	PqSendPointer = PqRecvPointer = PqRecvLength = 0;
 	DoingCopyOut = false;
-	if (getenv("LIBPQ_DEBUG"))
-		debug_port = stderr;
+	on_proc_exit(pq_close, 0);
 }
 
-/* --------------------------------
- *		pq_getport - return the PGPORT setting
- * --------------------------------
- */
-int
-pq_getport(void)
-{
-	char	   *envport = getenv("PGPORT");
-
-	if (envport)
-		return atoi(envport);
-	return atoi(DEF_PGPORT);
-}
 
 /* --------------------------------
  *		pq_close - shutdown libpq at backend exit
@@ -140,7 +132,7 @@ pq_getport(void)
  * don't crash during exit...
  * --------------------------------
  */
-void
+static void
 pq_close(void)
 {
 	if (MyProcPort != NULL)
@@ -168,7 +160,7 @@ static char sock_path[MAXPGPATH];
  * If a Unix socket is used for communication, explicitly close it.
  */
 static void
-StreamDoUnlink()
+StreamDoUnlink(void)
 {
 	Assert(sock_path[0]);
 	unlink(sock_path);
@@ -177,27 +169,23 @@ StreamDoUnlink()
 /*
  * StreamServerPort -- open a sock stream "listening" port.
  *
- * This initializes the Postmaster's connection-accepting port.
+ * This initializes the Postmaster's connection-accepting port *fdP.
  *
  * RETURNS: STATUS_OK or STATUS_ERROR
  */
 
 int
-StreamServerPort(char *hostName, unsigned short portName, int *fdP)
+StreamServerPort(int family, char *hostName, unsigned short portNumber,
+				 char *unixSocketName, int *fdP)
 {
 	SockAddr	saddr;
 	int			fd,
-				err,
-				family;
-	size_t		len;
+				err;
+	int			maxconn;
+	size_t		len = 0;
 	int			one = 1;
 
-#ifdef HAVE_FCNTL_SETLK
-	int			lock_fd;
-
-#endif
-
-	family = ((hostName != NULL) ? AF_INET : AF_UNIX);
+	Assert(family == AF_INET || family == AF_UNIX);
 
 	if ((fd = socket(family, SOCK_STREAM, 0)) < 0)
 	{
@@ -225,51 +213,62 @@ StreamServerPort(char *hostName, unsigned short portName, int *fdP)
 
 	MemSet((char *) &saddr, 0, sizeof(saddr));
 	saddr.sa.sa_family = family;
+
+#ifdef HAVE_UNIX_SOCKETS
 	if (family == AF_UNIX)
 	{
-		len = UNIXSOCK_PATH(saddr.un, portName);
+		UNIXSOCK_PATH(saddr.un, portNumber, unixSocketName);
+		len = UNIXSOCK_LEN(saddr.un);
 		strcpy(sock_path, saddr.un.sun_path);
 
 		/*
-		 * If the socket exists but nobody has an advisory lock on it we
-		 * can safely delete the file.
+		 * Grab an interlock file associated with the socket file.
 		 */
-#ifdef HAVE_FCNTL_SETLK
-#ifndef __CYGWIN32__
-		if ((lock_fd = open(sock_path, O_WRONLY | O_NONBLOCK, 0666)) >= 0)
-#else
-		if ((lock_fd = open(sock_path, O_WRONLY | O_NONBLOCK | O_BINARY, 0666)) >= 0)
-#endif
-		{
-			struct flock lck;
+		if (!CreateSocketLockFile(sock_path, true))
+			return STATUS_ERROR;
 
-			lck.l_whence = SEEK_SET;
-			lck.l_start = lck.l_len = 0;
-			lck.l_type = F_WRLCK;
-			if (fcntl(lock_fd, F_SETLK, &lck) == 0)
-			{
-				TPRINTF(TRACE_VERBOSE, "flock on %s, deleting", sock_path);
-				unlink(sock_path);
-			}
-			else
-				TPRINTF(TRACE_VERBOSE, "flock failed for %s", sock_path);
-			close(lock_fd);
-		}
-#endif	 /* HAVE_FCNTL_SETLK */
+		/*
+		 * Once we have the interlock, we can safely delete any
+		 * pre-existing socket file to avoid failure at bind() time.
+		 */
+		unlink(sock_path);
 	}
-	else
+#endif   /* HAVE_UNIX_SOCKETS */
+
+	if (family == AF_INET)
 	{
-		saddr.in.sin_addr.s_addr = htonl(INADDR_ANY);
-		saddr.in.sin_port = htons(portName);
+		/* TCP/IP socket */
+		if (hostName[0] == '\0')
+			saddr.in.sin_addr.s_addr = htonl(INADDR_ANY);
+		else
+		{
+			struct hostent *hp;
+
+			hp = gethostbyname(hostName);
+			if ((hp == NULL) || (hp->h_addrtype != AF_INET))
+			{
+				snprintf(PQerrormsg, PQERRORMSG_LENGTH,
+				   "FATAL: StreamServerPort: gethostbyname(%s) failed\n",
+						 hostName);
+				fputs(PQerrormsg, stderr);
+				pqdebug("%s", PQerrormsg);
+				return STATUS_ERROR;
+			}
+			memmove((char *) &(saddr.in.sin_addr), (char *) hp->h_addr,
+					hp->h_length);
+		}
+
+		saddr.in.sin_port = htons(portNumber);
 		len = sizeof(struct sockaddr_in);
 	}
-	err = bind(fd, &saddr.sa, len);
+
+	err = bind(fd, (struct sockaddr *) & saddr.sa, len);
 	if (err < 0)
 	{
 		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 "FATAL: StreamServerPort: bind() failed: %s\n"
-			   "\tIs another postmaster already running on that port?\n",
-				 strerror(errno));
+				 "\tIs another postmaster already running on port %d?\n",
+				 strerror(errno), (int) portNumber);
 		if (family == AF_UNIX)
 			snprintf(PQerrormsg + strlen(PQerrormsg),
 					 PQERRORMSG_LENGTH - strlen(PQerrormsg),
@@ -284,45 +283,92 @@ StreamServerPort(char *hostName, unsigned short portName, int *fdP)
 		return STATUS_ERROR;
 	}
 
+#ifdef HAVE_UNIX_SOCKETS
 	if (family == AF_UNIX)
 	{
-		on_proc_exit(StreamDoUnlink, NULL);
+		/* Arrange to unlink the socket file at exit */
+		on_proc_exit(StreamDoUnlink, 0);
 
 		/*
-		 * Open the socket file and get an advisory lock on it. The
-		 * lock_fd is left open to keep the lock.
+		 * Fix socket ownership/permission if requested.  Note we must do
+		 * this before we listen() to avoid a window where unwanted
+		 * connections could get accepted.
 		 */
-#ifdef HAVE_FCNTL_SETLK
-#ifndef __CYGWIN32__
-		if ((lock_fd = open(sock_path, O_WRONLY | O_NONBLOCK, 0666)) >= 0)
-#else
-		if ((lock_fd = open(sock_path, O_WRONLY | O_NONBLOCK | O_BINARY, 0666)) >= 0)
-#endif
+		Assert(Unix_socket_group);
+		if (Unix_socket_group[0] != '\0')
 		{
-			struct flock lck;
+			char	   *endptr;
+			unsigned long int val;
+			gid_t		gid;
 
-			lck.l_whence = SEEK_SET;
-			lck.l_start = lck.l_len = 0;
-			lck.l_type = F_WRLCK;
-			if (fcntl(lock_fd, F_SETLK, &lck) != 0)
-				TPRINTF(TRACE_VERBOSE, "flock error for %s", sock_path);
+			val = strtoul(Unix_socket_group, &endptr, 10);
+			if (*endptr == '\0')
+			{
+				/* numeric group id */
+				gid = val;
+			}
+			else
+			{
+				/* convert group name to id */
+				struct group *gr;
+
+				gr = getgrnam(Unix_socket_group);
+				if (!gr)
+				{
+					snprintf(PQerrormsg, PQERRORMSG_LENGTH,
+							 "FATAL:  no such group '%s'\n",
+							 Unix_socket_group);
+					fputs(PQerrormsg, stderr);
+					pqdebug("%s", PQerrormsg);
+					return STATUS_ERROR;
+				}
+				gid = gr->gr_gid;
+			}
+			if (chown(sock_path, -1, gid) == -1)
+			{
+				snprintf(PQerrormsg, PQERRORMSG_LENGTH,
+						 "FATAL:  could not set group of %s: %s\n",
+						 sock_path, strerror(errno));
+				fputs(PQerrormsg, stderr);
+				pqdebug("%s", PQerrormsg);
+				return STATUS_ERROR;
+			}
 		}
-#endif	 /* HAVE_FCNTL_SETLK */
-	}
 
-	listen(fd, SOMAXCONN);
+		if (chmod(sock_path, Unix_socket_permissions) == -1)
+		{
+			snprintf(PQerrormsg, PQERRORMSG_LENGTH,
+					 "FATAL:  could not set permissions on %s: %s\n",
+					 sock_path, strerror(errno));
+			fputs(PQerrormsg, stderr);
+			pqdebug("%s", PQerrormsg);
+			return STATUS_ERROR;
+		}
+	}
+#endif   /* HAVE_UNIX_SOCKETS */
 
 	/*
-	 * MS: I took this code from Dillon's version.  It makes the listening
-	 * port non-blocking.  That is not necessary (and may tickle kernel
-	 * bugs).
-	 *
-	 * fcntl(fd, F_SETFD, 1); fcntl(fd, F_SETFL, FNDELAY);
+	 * Select appropriate accept-queue length limit.  PG_SOMAXCONN is only
+	 * intended to provide a clamp on the request on platforms where an
+	 * overly large request provokes a kernel error (are there any?).
 	 */
+	maxconn = MaxBackends * 2;
+	if (maxconn > PG_SOMAXCONN)
+		maxconn = PG_SOMAXCONN;
+
+	err = listen(fd, maxconn);
+	if (err < 0)
+	{
+		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
+				 "FATAL: StreamServerPort: listen() failed: %s\n",
+				 strerror(errno));
+		fputs(PQerrormsg, stderr);
+		pqdebug("%s", PQerrormsg);
+		return STATUS_ERROR;
+	}
 
 	*fdP = fd;
-	if (family == AF_UNIX)
-		chmod(sock_path, 0777);
+
 	return STATUS_OK;
 }
 
@@ -343,7 +389,7 @@ StreamServerPort(char *hostName, unsigned short portName, int *fdP)
 int
 StreamConnection(int server_fd, Port *port)
 {
-	SOCKET_SIZE_TYPE addrlen;
+	ACCEPT_TYPE_ARG3 addrlen;
 
 	/* accept connection (and fill in the client (remote) address) */
 	addrlen = sizeof(port->raddr);
@@ -355,6 +401,16 @@ StreamConnection(int server_fd, Port *port)
 		return STATUS_ERROR;
 	}
 
+#ifdef SCO_ACCEPT_BUG
+
+	/*
+	 * UnixWare 7+ and OpenServer 5.0.4 are known to have this bug, but it
+	 * shouldn't hurt to catch it for all versions of those platforms.
+	 */
+	if (port->raddr.sa.sa_family == 0)
+		port->raddr.sa.sa_family = AF_UNIX;
+#endif
+
 	/* fill in the server (local) address */
 	addrlen = sizeof(port->laddr);
 	if (getsockname(port->sock, (struct sockaddr *) & port->laddr,
@@ -370,22 +426,19 @@ StreamConnection(int server_fd, Port *port)
 		int			on = 1;
 
 		if (setsockopt(port->sock, IPPROTO_TCP, TCP_NODELAY,
-					   &on, sizeof(on)) < 0)
+					   (char *) &on, sizeof(on)) < 0)
 		{
 			perror("postmaster: StreamConnection: setsockopt(TCP_NODELAY)");
 			return STATUS_ERROR;
 		}
 		if (setsockopt(port->sock, SOL_SOCKET, SO_KEEPALIVE,
-					   &on, sizeof(on)) < 0)
+					   (char *) &on, sizeof(on)) < 0)
 		{
 			perror("postmaster: StreamConnection: setsockopt(SO_KEEPALIVE)");
 			return STATUS_ERROR;
 		}
 	}
 
-	/* reset to non-blocking */
-	fcntl(port->sock, F_SETFL, 1);
-
 	return STATUS_OK;
 }
 
@@ -451,19 +504,18 @@ pq_recvbuf(void)
 				continue;		/* Ok if interrupted */
 
 			/*
-			 * We would like to use elog() here, but dare not because elog
-			 * tries to write to the client, which will cause problems if
-			 * we have a hard communications failure ... So just write the
-			 * message to the postmaster log.
+			 * Careful: an elog() that tries to write to the client
+			 * would cause recursion to here, leading to stack overflow
+			 * and core dump!  This message must go *only* to the postmaster
+			 * log.  elog(DEBUG) is presently safe.
 			 */
-			fprintf(stderr, "pq_recvbuf: recv() failed: %s\n",
-					strerror(errno));
+			elog(DEBUG, "pq_recvbuf: recv() failed: %m");
 			return EOF;
 		}
 		if (r == 0)
 		{
-			/* as above, elog not safe */
-			fprintf(stderr, "pq_recvbuf: unexpected EOF on client connection\n");
+			/* as above, only write to postmaster log */
+			elog(DEBUG, "pq_recvbuf: unexpected EOF on client connection");
 			return EOF;
 		}
 		/* r contains number of bytes read, so just incr length */
@@ -476,7 +528,7 @@ pq_recvbuf(void)
  *		pq_getbyte	- get a single byte from connection, or return EOF
  * --------------------------------
  */
-static int
+int
 pq_getbyte(void)
 {
 	while (PqRecvPointer >= PqRecvLength)
@@ -558,7 +610,7 @@ pq_getstring(StringInfo s)
 
 	/* Read until we get the terminating '\0' */
 	while ((c = pq_getbyte()) != EOF && c != '\0')
-		appendStringInfoChar(s, c);
+		appendStringInfoCharMacro(s, c);
 
 	if (c == EOF)
 		return EOF;
@@ -603,6 +655,8 @@ pq_putbytes(const char *s, size_t len)
 int
 pq_flush(void)
 {
+	static int last_reported_send_errno = 0;
+
 	unsigned char *bufptr = PqSendBuffer;
 	unsigned char *bufend = PqSendBuffer + PqSendPointer;
 
@@ -623,12 +677,20 @@ pq_flush(void)
 				continue;		/* Ok if we were interrupted */
 
 			/*
-			 * We would like to use elog() here, but cannot because elog
-			 * tries to write to the client, which would cause a recursive
-			 * flush attempt!  So just write it out to the postmaster log.
+			 * Careful: an elog() that tries to write to the client
+			 * would cause recursion to here, leading to stack overflow
+			 * and core dump!  This message must go *only* to the postmaster
+			 * log.  elog(DEBUG) is presently safe.
+			 *
+			 * If a client disconnects while we're in the midst of output,
+			 * we might write quite a bit of data before we get to a safe
+			 * query abort point.  So, suppress duplicate log messages.
 			 */
-			fprintf(stderr, "pq_flush: send() failed: %s\n",
-					strerror(errno));
+			if (errno != last_reported_send_errno)
+			{
+				last_reported_send_errno = errno;
+				elog(DEBUG, "pq_flush: send() failed: %m");
+			}
 
 			/*
 			 * We drop the buffered data anyway so that processing can
@@ -637,13 +699,40 @@ pq_flush(void)
 			PqSendPointer = 0;
 			return EOF;
 		}
+
+		last_reported_send_errno = 0; /* reset after any successful send */
 		bufptr += r;
 	}
+
 	PqSendPointer = 0;
 	return 0;
 }
 
 
+/*
+ * Return EOF if the connection has been broken, else 0.
+ */
+int
+pq_eof(void)
+{
+	char		x;
+	int			res;
+
+	res = recv(MyProcPort->sock, &x, 1, MSG_PEEK);
+
+	if (res < 0)
+	{
+		/* can log to postmaster log only */
+		elog(DEBUG, "pq_eof: recv() failed: %m");
+		return EOF;
+	}
+	if (res == 0)
+		return EOF;
+	else
+		return 0;
+}
+
+
 /* --------------------------------
  * Message-level I/O routines begin here.
  *
