@@ -8,13 +8,16 @@
  *
  *
  * IDENTIFICATION
- *	  $Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/optimizer/path/allpaths.c,v 1.72 2001/03/22 03:59:34 momjian Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/optimizer/path/allpaths.c,v 1.83 2001/12/10 22:54:12 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
 
 #include "postgres.h"
 
+#ifdef OPTIMIZER_DEBUG
+#include "nodes/print.h"
+#endif
 #include "optimizer/clauses.h"
 #include "optimizer/cost.h"
 #include "optimizer/geqo.h"
@@ -35,16 +38,13 @@ static void set_base_rel_pathlists(Query *root);
 static void set_plain_rel_pathlist(Query *root, RelOptInfo *rel,
 					   RangeTblEntry *rte);
 static void set_inherited_rel_pathlist(Query *root, RelOptInfo *rel,
-						   RangeTblEntry *rte,
+						   Index rti, RangeTblEntry *rte,
 						   List *inheritlist);
+static void set_subquery_pathlist(Query *root, RelOptInfo *rel,
+					  Index rti, RangeTblEntry *rte);
 static RelOptInfo *make_one_rel_by_joins(Query *root, int levels_needed,
 					  List *initial_rels);
 
-#ifdef OPTIMIZER_DEBUG
-static void debug_print_rel(Query *root, RelOptInfo *rel);
-
-#endif
-
 
 /*
  * make_one_rel
@@ -69,7 +69,7 @@ make_one_rel(Query *root)
 	rel = make_fromexpr_rel(root, root->jointree);
 
 	/*
-	 * The result should join all the query's rels.
+	 * The result should join all the query's base rels.
 	 */
 	Assert(length(rel->relids) == length(root->base_rel_list));
 
@@ -101,105 +101,23 @@ set_base_rel_pathlists(Query *root)
 		if (rel->issubquery)
 		{
 			/* Subquery --- generate a separate plan for it */
-
-			/*
-			 * If there are any restriction clauses that have been
-			 * attached to the subquery relation, consider pushing them
-			 * down to become HAVING quals of the subquery itself.	(Not
-			 * WHERE clauses, since they may refer to subquery outputs
-			 * that are aggregate results. But planner.c will transfer
-			 * them into the subquery's WHERE if they do not.)  This
-			 * transformation is useful because it may allow us to
-			 * generate a better plan for the subquery than evaluating all
-			 * the subquery output rows and then filtering them.
-			 *
-			 * Currently, we do not push down clauses that contain
-			 * subselects, mainly because I'm not sure it will work
-			 * correctly (the subplan hasn't yet transformed sublinks to
-			 * subselects). Also, if the subquery contains set ops
-			 * (UNION/INTERSECT/EXCEPT) we do not push down any qual
-			 * clauses, since the planner doesn't support quals at the top
-			 * level of a setop.  (With suitable analysis we could try to
-			 * push the quals down into the component queries of the
-			 * setop, but getting it right is not trivial.)
-			 * Non-pushed-down clauses will get evaluated as qpquals of
-			 * the SubqueryScan node.
-			 *
-			 * XXX Are there any cases where we want to make a policy
-			 * decision not to push down, because it'd result in a worse
-			 * plan?
-			 */
-			if (rte->subquery->setOperations == NULL)
-			{
-				/* OK to consider pushing down individual quals */
-				List	   *upperrestrictlist = NIL;
-				List	   *lst;
-
-				foreach(lst, rel->baserestrictinfo)
-				{
-					RestrictInfo *rinfo = (RestrictInfo *) lfirst(lst);
-					Node	   *clause = (Node *) rinfo->clause;
-
-					if (contain_subplans(clause))
-					{
-						/* Keep it in the upper query */
-						upperrestrictlist = lappend(upperrestrictlist, rinfo);
-					}
-					else
-					{
-
-						/*
-						 * We need to replace Vars in the clause (which
-						 * must refer to outputs of the subquery) with
-						 * copies of the subquery's targetlist
-						 * expressions.  Note that at this point, any
-						 * uplevel Vars in the clause should have been
-						 * replaced with Params, so they need no work.
-						 */
-						clause = ResolveNew(clause, rti, 0,
-											rte->subquery->targetList,
-											CMD_SELECT, 0);
-						rte->subquery->havingQual =
-							make_and_qual(rte->subquery->havingQual,
-										  clause);
-
-						/*
-						 * We need not change the subquery's hasAggs or
-						 * hasSublinks flags, since we can't be pushing
-						 * down any aggregates that weren't there before,
-						 * and we don't push down subselects at all.
-						 */
-					}
-				}
-				rel->baserestrictinfo = upperrestrictlist;
-			}
-
-			/* Generate the plan for the subquery */
-			rel->subplan = subquery_planner(rte->subquery,
-											-1.0 /* default case */ );
-
-			/* Copy number of output rows from subplan */
-			rel->tuples = rel->subplan->plan_rows;
-
-			/* Mark rel with estimated output rows, width, etc */
-			set_baserel_size_estimates(root, rel);
-
-			/* Generate appropriate path */
-			add_path(rel, create_subqueryscan_path(rel));
-
-			/* Select cheapest path (pretty easy in this case...) */
-			set_cheapest(rel);
+			set_subquery_pathlist(root, rel, rti, rte);
 		}
-		else if ((inheritlist = expand_inherted_rtentry(root, rti)) != NIL)
+		else if ((inheritlist = expand_inherted_rtentry(root, rti, true))
+				 != NIL)
 		{
 			/* Relation is root of an inheritance tree, process specially */
-			set_inherited_rel_pathlist(root, rel, rte, inheritlist);
+			set_inherited_rel_pathlist(root, rel, rti, rte, inheritlist);
 		}
 		else
 		{
 			/* Plain relation */
 			set_plain_rel_pathlist(root, rel, rte);
 		}
+
+#ifdef OPTIMIZER_DEBUG
+		debug_print_rel(root, rel);
+#endif
 	}
 }
 
@@ -210,8 +128,6 @@ set_base_rel_pathlists(Query *root)
 static void
 set_plain_rel_pathlist(Query *root, RelOptInfo *rel, RangeTblEntry *rte)
 {
-	List	   *indices = find_secondary_indexes(rte->relid);
-
 	/* Mark rel with estimated output rows, width, etc */
 	set_baserel_size_estimates(root, rel);
 
@@ -224,20 +140,16 @@ set_plain_rel_pathlist(Query *root, RelOptInfo *rel, RangeTblEntry *rte)
 	 */
 
 	/* Consider sequential scan */
-	add_path(rel, create_seqscan_path(rel));
+	add_path(rel, create_seqscan_path(root, rel));
 
 	/* Consider TID scans */
 	create_tidscan_paths(root, rel);
 
 	/* Consider index paths for both simple and OR index clauses */
-	create_index_paths(root, rel, indices);
+	create_index_paths(root, rel);
 
-	/*
-	 * Note: create_or_index_paths depends on create_index_paths to have
-	 * marked OR restriction clauses with relevant indices; this is why it
-	 * doesn't need to be given the list of indices.
-	 */
-	create_or_index_paths(root, rel, rel->baserestrictinfo);
+	/* create_index_paths must be done before create_or_index_paths */
+	create_or_index_paths(root, rel);
 
 	/* Now find the cheapest of the paths for this rel */
 	set_cheapest(rel);
@@ -248,14 +160,26 @@ set_plain_rel_pathlist(Query *root, RelOptInfo *rel, RangeTblEntry *rte)
  *	  Build access paths for a inheritance tree rooted at rel
  *
  * inheritlist is a list of RT indexes of all tables in the inheritance tree,
- * including the parent itself.  Note we will not come here unless there's
- * at least one child in addition to the parent.
+ * including a duplicate of the parent itself.	Note we will not come here
+ * unless there's at least one child in addition to the parent.
+ *
+ * NOTE: the passed-in rel and RTE will henceforth represent the appended
+ * result of the whole inheritance tree.  The members of inheritlist represent
+ * the individual tables --- in particular, the inheritlist member that is a
+ * duplicate of the parent RTE represents the parent table alone.
+ * We will generate plans to scan the individual tables that refer to
+ * the inheritlist RTEs, whereas Vars elsewhere in the plan tree that
+ * refer to the original RTE are taken to refer to the append output.
+ * In particular, this means we have separate RelOptInfos for the parent
+ * table and for the append output, which is a good thing because they're
+ * not the same size.
  */
 static void
-set_inherited_rel_pathlist(Query *root, RelOptInfo *rel, RangeTblEntry *rte,
+set_inherited_rel_pathlist(Query *root, RelOptInfo *rel,
+						   Index rti, RangeTblEntry *rte,
 						   List *inheritlist)
 {
-	int			parentRTindex = lfirsti(rel->relids);
+	int			parentRTindex = rti;
 	Oid			parentOID = rte->relid;
 	List	   *subpaths = NIL;
 	List	   *il;
@@ -268,7 +192,15 @@ set_inherited_rel_pathlist(Query *root, RelOptInfo *rel, RangeTblEntry *rte,
 		elog(ERROR, "SELECT FOR UPDATE is not supported for inherit queries");
 
 	/*
-	 * Recompute size estimates for whole inheritance tree
+	 * The executor will check the parent table's access permissions when
+	 * it examines the parent's inheritlist entry.  There's no need to
+	 * check twice, so turn off access check bits in the original RTE.
+	 */
+	rte->checkForRead = false;
+	rte->checkForWrite = false;
+
+	/*
+	 * Initialize to compute size estimates for whole inheritance tree
 	 */
 	rel->rows = 0;
 	rel->width = 0;
@@ -289,21 +221,17 @@ set_inherited_rel_pathlist(Query *root, RelOptInfo *rel, RangeTblEntry *rte,
 
 		/*
 		 * Make a RelOptInfo for the child so we can do planning.  Do NOT
-		 * attach the RelOptInfo to the query's base_rel_list, however.
-		 *
-		 * NOTE: when childRTindex == parentRTindex, we create a second
-		 * RelOptInfo for the same relation.  This RelOptInfo will
-		 * represent the parent table alone, whereas the original
-		 * RelOptInfo represents the union of the inheritance tree
-		 * members.
+		 * attach the RelOptInfo to the query's base_rel_list, however,
+		 * since the child is not part of the main join tree.  Instead,
+		 * the child RelOptInfo is added to other_rel_list.
 		 */
-		childrel = make_base_rel(root, childRTindex);
+		childrel = build_other_rel(root, childRTindex);
 
 		/*
 		 * Copy the parent's targetlist and restriction quals to the
-		 * child, with attribute-number adjustment if needed.  We don't
-		 * bother to copy the join quals, since we can't do any joining
-		 * here.
+		 * child, with attribute-number adjustment as needed.  We don't
+		 * bother to copy the join quals, since we can't do any joining of
+		 * the individual tables.
 		 */
 		childrel->targetlist = (List *)
 			adjust_inherited_attrs((Node *) rel->targetlist,
@@ -342,6 +270,120 @@ set_inherited_rel_pathlist(Query *root, RelOptInfo *rel, RangeTblEntry *rte,
 	set_cheapest(rel);
 }
 
+/*
+ * set_subquery_pathlist
+ *		Build the (single) access path for a subquery RTE
+ */
+static void
+set_subquery_pathlist(Query *root, RelOptInfo *rel,
+					  Index rti, RangeTblEntry *rte)
+{
+	Query	   *subquery = rte->subquery;
+
+	/*
+	 * If there are any restriction clauses that have been attached to the
+	 * subquery relation, consider pushing them down to become HAVING
+	 * quals of the subquery itself.  (Not WHERE clauses, since they may
+	 * refer to subquery outputs that are aggregate results.  But
+	 * planner.c will transfer them into the subquery's WHERE if they do
+	 * not.)  This transformation is useful because it may allow us to
+	 * generate a better plan for the subquery than evaluating all the
+	 * subquery output rows and then filtering them.
+	 *
+	 * There are several cases where we cannot push down clauses:
+	 *
+	 * 1. If the subquery contains set ops (UNION/INTERSECT/EXCEPT) we do not
+	 * push down any qual clauses, since the planner doesn't support quals
+	 * at the top level of a setop.  (With suitable analysis we could try
+	 * to push the quals down into the component queries of the setop, but
+	 * getting it right seems nontrivial.  Work on this later.)
+	 *
+	 * 2. If the subquery has a LIMIT clause or a DISTINCT ON clause, we must
+	 * not push down any quals, since that could change the set of rows
+	 * returned.  (Actually, we could push down quals into a DISTINCT ON
+	 * subquery if they refer only to DISTINCT-ed output columns, but
+	 * checking that seems more work than it's worth.  In any case, a
+	 * plain DISTINCT is safe to push down past.)
+	 *
+	 * 3. If the subquery has any ITER nodes (ie, functions returning sets)
+	 * in its target list, we do not push down any quals, since the quals
+	 * might refer to those tlist items, which would mean we'd introduce
+	 * functions-returning-sets into the subquery's WHERE/HAVING quals.
+	 * (It'd be sufficient to not push down quals that refer to those
+	 * particular tlist items, but that's much clumsier to check.)
+	 *
+	 * 4. We do not push down clauses that contain subselects, mainly because
+	 * I'm not sure it will work correctly (the subplan hasn't yet
+	 * transformed sublinks to subselects).
+	 *
+	 * Non-pushed-down clauses will get evaluated as qpquals of the
+	 * SubqueryScan node.
+	 *
+	 * XXX Are there any cases where we want to make a policy decision not to
+	 * push down, because it'd result in a worse plan?
+	 */
+	if (subquery->setOperations == NULL &&
+		subquery->limitOffset == NULL &&
+		subquery->limitCount == NULL &&
+		!has_distinct_on_clause(subquery) &&
+		!contain_iter_clause((Node *) subquery->targetList))
+	{
+		/* OK to consider pushing down individual quals */
+		List	   *upperrestrictlist = NIL;
+		List	   *lst;
+
+		foreach(lst, rel->baserestrictinfo)
+		{
+			RestrictInfo *rinfo = (RestrictInfo *) lfirst(lst);
+			Node	   *clause = (Node *) rinfo->clause;
+
+			if (contain_subplans(clause))
+			{
+				/* Keep it in the upper query */
+				upperrestrictlist = lappend(upperrestrictlist, rinfo);
+			}
+			else
+			{
+				/*
+				 * We need to replace Vars in the clause (which must refer
+				 * to outputs of the subquery) with copies of the
+				 * subquery's targetlist expressions.  Note that at this
+				 * point, any uplevel Vars in the clause should have been
+				 * replaced with Params, so they need no work.
+				 */
+				clause = ResolveNew(clause, rti, 0,
+									subquery->targetList,
+									CMD_SELECT, 0);
+				subquery->havingQual = make_and_qual(subquery->havingQual,
+													 clause);
+
+				/*
+				 * We need not change the subquery's hasAggs or
+				 * hasSublinks flags, since we can't be pushing down any
+				 * aggregates that weren't there before, and we don't push
+				 * down subselects at all.
+				 */
+			}
+		}
+		rel->baserestrictinfo = upperrestrictlist;
+	}
+
+	/* Generate the plan for the subquery */
+	rel->subplan = subquery_planner(subquery,
+									-1.0 /* default case */ );
+
+	/* Copy number of output rows from subplan */
+	rel->tuples = rel->subplan->plan_rows;
+
+	/* Mark rel with estimated output rows, width, etc */
+	set_baserel_size_estimates(root, rel);
+
+	/* Generate appropriate path */
+	add_path(rel, create_subqueryscan_path(rel));
+
+	/* Select cheapest path (pretty easy in this case...) */
+	set_cheapest(rel);
+}
 
 /*
  * make_fromexpr_rel
@@ -379,7 +421,6 @@ make_fromexpr_rel(Query *root, FromExpr *from)
 
 	if (levels_needed == 1)
 	{
-
 		/*
 		 * Single jointree node, so we're done.
 		 */
@@ -387,7 +428,6 @@ make_fromexpr_rel(Query *root, FromExpr *from)
 	}
 	else
 	{
-
 		/*
 		 * Consider the different orders in which we could join the rels,
 		 * using either GEQO or regular optimizer.
@@ -489,10 +529,22 @@ make_one_rel_by_joins(Query *root, int levels_needed, List *initial_rels)
 #ifdef OPTIMIZER_DEBUG
 
 static void
-print_joinclauses(Query *root, List *clauses)
+print_relids(Relids relids)
+{
+	List	   *l;
+
+	foreach(l, relids)
+	{
+		printf("%d", lfirsti(l));
+		if (lnext(l))
+			printf(" ");
+	}
+}
+
+static void
+print_restrictclauses(Query *root, List *clauses)
 {
 	List	   *l;
-	extern void print_expr(Node *expr, List *rtable);	/* in print.c */
 
 	foreach(l, clauses)
 	{
@@ -500,21 +552,17 @@ print_joinclauses(Query *root, List *clauses)
 
 		print_expr((Node *) c->clause, root->rtable);
 		if (lnext(l))
-			printf(" ");
+			printf(", ");
 	}
 }
 
 static void
 print_path(Query *root, Path *path, int indent)
 {
-	char	   *ptype = NULL;
-	JoinPath   *jp;
-	bool		join = false;
+	const char *ptype;
+	bool		join;
 	int			i;
 
-	for (i = 0; i < indent; i++)
-		printf("\t");
-
 	switch (nodeTag(path))
 	{
 		case T_Path:
@@ -525,6 +573,10 @@ print_path(Query *root, Path *path, int indent)
 			ptype = "IdxScan";
 			join = false;
 			break;
+		case T_TidPath:
+			ptype = "TidScan";
+			join = false;
+			break;
 		case T_NestPath:
 			ptype = "Nestloop";
 			join = true;
@@ -538,89 +590,91 @@ print_path(Query *root, Path *path, int indent)
 			join = true;
 			break;
 		default:
+			ptype = "???Path";
+			join = false;
 			break;
 	}
+
+	for (i = 0; i < indent; i++)
+		printf("\t");
+	printf("%s(", ptype);
+	print_relids(path->parent->relids);
+	printf(") rows=%.0f cost=%.2f..%.2f\n",
+		   path->parent->rows, path->startup_cost, path->total_cost);
+
+	if (path->pathkeys)
+	{
+		for (i = 0; i < indent; i++)
+			printf("\t");
+		printf("  pathkeys: ");
+		print_pathkeys(path->pathkeys, root->rtable);
+	}
+
 	if (join)
 	{
-		jp = (JoinPath *) path;
+		JoinPath   *jp = (JoinPath *) path;
 
-		printf("%s rows=%.0f cost=%.2f..%.2f\n",
-			   ptype, path->parent->rows,
-			   path->startup_cost, path->total_cost);
+		for (i = 0; i < indent; i++)
+			printf("\t");
+		printf("  clauses: ");
+		print_restrictclauses(root, jp->joinrestrictinfo);
+		printf("\n");
 
-		if (path->pathkeys)
+		if (nodeTag(path) == T_MergePath)
 		{
-			for (i = 0; i < indent; i++)
-				printf("\t");
-			printf("  pathkeys=");
-			print_pathkeys(path->pathkeys, root->rtable);
-		}
+			MergePath  *mp = (MergePath *) path;
 
-		switch (nodeTag(path))
-		{
-			case T_MergePath:
-			case T_HashPath:
+			if (mp->outersortkeys || mp->innersortkeys)
+			{
 				for (i = 0; i < indent; i++)
 					printf("\t");
-				printf("  clauses=(");
-				print_joinclauses(root, jp->joinrestrictinfo);
-				printf(")\n");
-
-				if (nodeTag(path) == T_MergePath)
-				{
-					MergePath  *mp = (MergePath *) path;
-
-					if (mp->outersortkeys || mp->innersortkeys)
-					{
-						for (i = 0; i < indent; i++)
-							printf("\t");
-						printf("  sortouter=%d sortinner=%d\n",
-							   ((mp->outersortkeys) ? 1 : 0),
-							   ((mp->innersortkeys) ? 1 : 0));
-					}
-				}
-				break;
-			default:
-				break;
+				printf("  sortouter=%d sortinner=%d\n",
+					   ((mp->outersortkeys) ? 1 : 0),
+					   ((mp->innersortkeys) ? 1 : 0));
+			}
 		}
+
 		print_path(root, jp->outerjoinpath, indent + 1);
 		print_path(root, jp->innerjoinpath, indent + 1);
 	}
-	else
-	{
-		int			relid = lfirsti(path->parent->relids);
-
-		printf("%s(%d) rows=%.0f cost=%.2f..%.2f\n",
-			   ptype, relid, path->parent->rows,
-			   path->startup_cost, path->total_cost);
-
-		if (path->pathkeys)
-		{
-			for (i = 0; i < indent; i++)
-				printf("\t");
-			printf("  pathkeys=");
-			print_pathkeys(path->pathkeys, root->rtable);
-		}
-	}
 }
 
-static void
+void
 debug_print_rel(Query *root, RelOptInfo *rel)
 {
 	List	   *l;
 
-	printf("(");
-	foreach(l, rel->relids)
-		printf("%d ", lfirsti(l));
+	printf("RELOPTINFO (");
+	print_relids(rel->relids);
 	printf("): rows=%.0f width=%d\n", rel->rows, rel->width);
 
+	if (rel->baserestrictinfo)
+	{
+		printf("\tbaserestrictinfo: ");
+		print_restrictclauses(root, rel->baserestrictinfo);
+		printf("\n");
+	}
+
+	foreach(l, rel->joininfo)
+	{
+		JoinInfo   *j = (JoinInfo *) lfirst(l);
+
+		printf("\tjoininfo (");
+		print_relids(j->unjoined_relids);
+		printf("): ");
+		print_restrictclauses(root, j->jinfo_restrictinfo);
+		printf("\n");
+	}
+
 	printf("\tpath list:\n");
 	foreach(l, rel->pathlist)
 		print_path(root, lfirst(l), 1);
-	printf("\tcheapest startup path:\n");
+	printf("\n\tcheapest startup path:\n");
 	print_path(root, rel->cheapest_startup_path, 1);
-	printf("\tcheapest total path:\n");
+	printf("\n\tcheapest total path:\n");
 	print_path(root, rel->cheapest_total_path, 1);
+	printf("\n");
+	fflush(stdout);
 }
 
-#endif	 /* OPTIMIZER_DEBUG */
+#endif   /* OPTIMIZER_DEBUG */
