@@ -3,25 +3,24 @@
  * pathnode.c
  *	  Routines to manipulate pathlists and create path nodes
  *
- * Portions Copyright (c) 1996-2000, PostgreSQL, Inc
+ * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /usr/local/cvsroot/pgsql/src/backend/optimizer/util/pathnode.c,v 1.63 2000/04/12 17:15:24 momjian Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/optimizer/util/pathnode.c,v 1.76 2001/10/25 05:49:34 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
-#include <math.h>
-
 #include "postgres.h"
 
+#include <math.h>
+
+#include "nodes/plannodes.h"
 #include "optimizer/cost.h"
 #include "optimizer/pathnode.h"
 #include "optimizer/paths.h"
-#include "optimizer/plancat.h"
 #include "optimizer/restrictinfo.h"
-#include "parser/parsetree.h"
 
 
 /*****************************************************************************
@@ -121,7 +120,9 @@ set_cheapest(RelOptInfo *parent_rel)
 	Path	   *cheapest_total_path;
 
 	Assert(IsA(parent_rel, RelOptInfo));
-	Assert(pathlist != NIL);
+
+	if (pathlist == NIL)
+		elog(ERROR, "Unable to devise a query plan for the given query");
 
 	cheapest_startup_path = cheapest_total_path = (Path *) lfirst(pathlist);
 
@@ -160,11 +161,17 @@ set_cheapest(RelOptInfo *parent_rel)
  *	  pathlist any old paths that are dominated by new_path --- that is,
  *	  new_path is both cheaper and at least as well ordered.
  *
+ *	  The pathlist is kept sorted by TOTAL_COST metric, with cheaper paths
+ *	  at the front.  No code depends on that for correctness; it's simply
+ *	  a speed hack within this routine.  Doing it that way makes it more
+ *	  likely that we will reject an inferior path after a few comparisons,
+ *	  rather than many comparisons.
+ *
  *	  NOTE: discarded Path objects are immediately pfree'd to reduce planner
  *	  memory consumption.  We dare not try to free the substructure of a Path,
  *	  since much of it may be shared with other Paths or the query tree itself;
  *	  but just recycling discarded Path nodes is a very useful savings in
- *	  a large join tree.
+ *	  a large join tree.  We can recycle the List nodes of pathlist, too.
  *
  * 'parent_rel' is the relation entry to which the path corresponds.
  * 'new_path' is a potential path for parent_rel.
@@ -176,6 +183,7 @@ add_path(RelOptInfo *parent_rel, Path *new_path)
 {
 	bool		accept_new = true;		/* unless we find a superior old
 										 * path */
+	List	   *insert_after = NIL;		/* where to insert new item */
 	List	   *p1_prev = NIL;
 	List	   *p1;
 
@@ -184,7 +192,8 @@ add_path(RelOptInfo *parent_rel, Path *new_path)
 	 * possible for more than one old path to be tossed out because
 	 * new_path dominates it.
 	 */
-	foreach(p1, parent_rel->pathlist)
+	p1 = parent_rel->pathlist;	/* cannot use foreach here */
+	while (p1 != NIL)
 	{
 		Path	   *old_path = (Path *) lfirst(p1);
 		bool		remove_old = false; /* unless new proves superior */
@@ -196,13 +205,14 @@ add_path(RelOptInfo *parent_rel, Path *new_path)
 		 * If the two paths compare differently for startup and total
 		 * cost, then we want to keep both, and we can skip the (much
 		 * slower) comparison of pathkeys.	If they compare the same,
-		 * proceed with the pathkeys comparison.  Note this test relies on
-		 * the fact that compare_path_costs will only return 0 if both
+		 * proceed with the pathkeys comparison.  Note: this test relies
+		 * on the fact that compare_path_costs will only return 0 if both
 		 * costs are equal (and, therefore, there's no need to call it
 		 * twice in that case).
 		 */
 		if (costcmp == 0 ||
-		 costcmp == compare_path_costs(new_path, old_path, STARTUP_COST))
+			costcmp == compare_path_costs(new_path, old_path,
+										  STARTUP_COST))
 		{
 			switch (compare_pathkeys(new_path->pathkeys, old_path->pathkeys))
 			{
@@ -233,14 +243,24 @@ add_path(RelOptInfo *parent_rel, Path *new_path)
 		 */
 		if (remove_old && parent_rel->pruneable)
 		{
+			List	   *p1_next = lnext(p1);
+
 			if (p1_prev)
-				lnext(p1_prev) = lnext(p1);
+				lnext(p1_prev) = p1_next;
 			else
-				parent_rel->pathlist = lnext(p1);
+				parent_rel->pathlist = p1_next;
 			pfree(old_path);
+			pfree(p1);			/* this is why we can't use foreach */
+			p1 = p1_next;
 		}
 		else
+		{
+			/* new belongs after this old path if it has cost >= old's */
+			if (costcmp >= 0)
+				insert_after = p1;
 			p1_prev = p1;
+			p1 = lnext(p1);
+		}
 
 		/*
 		 * If we found an old path that dominates new_path, we can quit
@@ -253,12 +273,15 @@ add_path(RelOptInfo *parent_rel, Path *new_path)
 
 	if (accept_new)
 	{
-		/* Accept the path */
-		parent_rel->pathlist = lcons(new_path, parent_rel->pathlist);
+		/* Accept the new path: insert it at proper place in pathlist */
+		if (insert_after)
+			lnext(insert_after) = lcons(new_path, lnext(insert_after));
+		else
+			parent_rel->pathlist = lcons(new_path, parent_rel->pathlist);
 	}
 	else
 	{
-		/* Reject and recycle the path */
+		/* Reject and recycle the new path */
 		pfree(new_path);
 	}
 }
@@ -272,10 +295,9 @@ add_path(RelOptInfo *parent_rel, Path *new_path)
  * create_seqscan_path
  *	  Creates a path corresponding to a sequential scan, returning the
  *	  pathnode.
- *
  */
 Path *
-create_seqscan_path(RelOptInfo *rel)
+create_seqscan_path(Query *root, RelOptInfo *rel)
 {
 	Path	   *pathnode = makeNode(Path);
 
@@ -283,7 +305,7 @@ create_seqscan_path(RelOptInfo *rel)
 	pathnode->parent = rel;
 	pathnode->pathkeys = NIL;	/* seqscan has unordered result */
 
-	cost_seqscan(pathnode, rel);
+	cost_seqscan(pathnode, root, rel);
 
 	return pathnode;
 }
@@ -296,18 +318,19 @@ create_seqscan_path(RelOptInfo *rel)
  * 'index' is an index on 'rel'
  * 'restriction_clauses' is a list of RestrictInfo nodes
  *			to be used as index qual conditions in the scan.
+ * 'pathkeys' describes the ordering of the path.
  * 'indexscandir' is ForwardScanDirection or BackwardScanDirection
- *			if the caller expects a specific scan direction,
- *			or NoMovementScanDirection if the caller is willing to accept
+ *			for an ordered index, or NoMovementScanDirection for
  *			an unordered index.
  *
  * Returns the new path node.
  */
-IndexPath  *
+IndexPath *
 create_index_path(Query *root,
 				  RelOptInfo *rel,
 				  IndexOptInfo *index,
 				  List *restriction_clauses,
+				  List *pathkeys,
 				  ScanDirection indexscandir)
 {
 	IndexPath  *pathnode = makeNode(IndexPath);
@@ -315,25 +338,7 @@ create_index_path(Query *root,
 
 	pathnode->path.pathtype = T_IndexScan;
 	pathnode->path.parent = rel;
-
-	pathnode->path.pathkeys = build_index_pathkeys(root, rel, index,
-												   indexscandir);
-	if (pathnode->path.pathkeys == NIL)
-	{
-		/* No ordering available from index, is that OK? */
-		if (!ScanDirectionIsNoMovement(indexscandir))
-			elog(ERROR, "create_index_path: failed to create ordered index scan");
-	}
-	else
-	{
-
-		/*
-		 * The index is ordered, and build_index_pathkeys defaulted to
-		 * forward scan, so make sure we mark the pathnode properly.
-		 */
-		if (ScanDirectionIsNoMovement(indexscandir))
-			indexscandir = ForwardScanDirection;
-	}
+	pathnode->path.pathkeys = pathkeys;
 
 	indexquals = get_actual_clauses(restriction_clauses);
 	/* expand special operators to indexquals the executor can handle */
@@ -341,10 +346,10 @@ create_index_path(Query *root,
 
 	/*
 	 * We are making a pathnode for a single-scan indexscan; therefore,
-	 * both indexid and indexqual should be single-element lists.
+	 * both indexinfo and indexqual should be single-element lists.
 	 */
-	pathnode->indexid = lconsi(index->indexoid, NIL);
-	pathnode->indexqual = lcons(indexquals, NIL);
+	pathnode->indexinfo = makeList1(index);
+	pathnode->indexqual = makeList1(indexquals);
 
 	pathnode->indexscandir = indexscandir;
 
@@ -354,8 +359,16 @@ create_index_path(Query *root,
 	 * number of rows is the same as the parent rel's estimate.
 	 */
 	pathnode->joinrelids = NIL; /* no join clauses here */
+	pathnode->alljoinquals = false;
 	pathnode->rows = rel->rows;
 
+	/*
+	 * Not sure if this is necessary, but it should help if the statistics
+	 * are too far off
+	 */
+	if (index->indpred && index->tuples < pathnode->rows)
+		pathnode->rows = index->tuples;
+
 	cost_index(&pathnode->path, root, rel, index, indexquals, false);
 
 	return pathnode;
@@ -365,10 +378,9 @@ create_index_path(Query *root,
  * create_tidscan_path
  *	  Creates a path corresponding to a tid_direct scan, returning the
  *	  pathnode.
- *
  */
-TidPath    *
-create_tidscan_path(RelOptInfo *rel, List *tideval)
+TidPath *
+create_tidscan_path(Query *root, RelOptInfo *rel, List *tideval)
 {
 	TidPath    *pathnode = makeNode(TidPath);
 
@@ -379,7 +391,7 @@ create_tidscan_path(RelOptInfo *rel, List *tideval)
 												 * necessary? */
 	pathnode->unjoined_relids = NIL;
 
-	cost_tidscan(&pathnode->path, rel, tideval);
+	cost_tidscan(&pathnode->path, root, rel, tideval);
 
 	/*
 	 * divide selectivity for each clause to get an equal selectivity as
@@ -389,22 +401,77 @@ create_tidscan_path(RelOptInfo *rel, List *tideval)
 	return pathnode;
 }
 
+/*
+ * create_append_path
+ *	  Creates a path corresponding to an Append plan, returning the
+ *	  pathnode.
+ *
+ */
+AppendPath *
+create_append_path(RelOptInfo *rel, List *subpaths)
+{
+	AppendPath *pathnode = makeNode(AppendPath);
+	List	   *l;
+
+	pathnode->path.pathtype = T_Append;
+	pathnode->path.parent = rel;
+	pathnode->path.pathkeys = NIL;		/* result is always considered
+										 * unsorted */
+	pathnode->subpaths = subpaths;
+
+	pathnode->path.startup_cost = 0;
+	pathnode->path.total_cost = 0;
+	foreach(l, subpaths)
+	{
+		Path	   *subpath = (Path *) lfirst(l);
+
+		if (l == subpaths)		/* first node? */
+			pathnode->path.startup_cost = subpath->startup_cost;
+		pathnode->path.total_cost += subpath->total_cost;
+	}
+
+	return pathnode;
+}
+
+/*
+ * create_subqueryscan_path
+ *	  Creates a path corresponding to a sequential scan of a subquery,
+ *	  returning the pathnode.
+ */
+Path *
+create_subqueryscan_path(RelOptInfo *rel)
+{
+	Path	   *pathnode = makeNode(Path);
+
+	pathnode->pathtype = T_SubqueryScan;
+	pathnode->parent = rel;
+	pathnode->pathkeys = NIL;	/* for now, assume unordered result */
+
+	/* just copy the subplan's cost estimates */
+	pathnode->startup_cost = rel->subplan->startup_cost;
+	pathnode->total_cost = rel->subplan->total_cost;
+
+	return pathnode;
+}
+
 /*
  * create_nestloop_path
  *	  Creates a pathnode corresponding to a nestloop join between two
  *	  relations.
  *
  * 'joinrel' is the join relation.
+ * 'jointype' is the type of join required
  * 'outer_path' is the outer path
  * 'inner_path' is the inner path
  * 'restrict_clauses' are the RestrictInfo nodes to apply at the join
  * 'pathkeys' are the path keys of the new join path
  *
  * Returns the resulting path node.
- *
  */
-NestPath   *
-create_nestloop_path(RelOptInfo *joinrel,
+NestPath *
+create_nestloop_path(Query *root,
+					 RelOptInfo *joinrel,
+					 JoinType jointype,
 					 Path *outer_path,
 					 Path *inner_path,
 					 List *restrict_clauses,
@@ -414,12 +481,14 @@ create_nestloop_path(RelOptInfo *joinrel,
 
 	pathnode->path.pathtype = T_NestLoop;
 	pathnode->path.parent = joinrel;
+	pathnode->jointype = jointype;
 	pathnode->outerjoinpath = outer_path;
 	pathnode->innerjoinpath = inner_path;
 	pathnode->joinrestrictinfo = restrict_clauses;
 	pathnode->path.pathkeys = pathkeys;
 
-	cost_nestloop(&pathnode->path, outer_path, inner_path, restrict_clauses);
+	cost_nestloop(&pathnode->path, root, outer_path, inner_path,
+				  restrict_clauses);
 
 	return pathnode;
 }
@@ -430,6 +499,7 @@ create_nestloop_path(RelOptInfo *joinrel,
  *	  two relations
  *
  * 'joinrel' is the join relation
+ * 'jointype' is the type of join required
  * 'outer_path' is the outer path
  * 'inner_path' is the inner path
  * 'restrict_clauses' are the RestrictInfo nodes to apply at the join
@@ -438,10 +508,11 @@ create_nestloop_path(RelOptInfo *joinrel,
  *		(this should be a subset of the restrict_clauses list)
  * 'outersortkeys' are the sort varkeys for the outer relation
  * 'innersortkeys' are the sort varkeys for the inner relation
- *
  */
-MergePath  *
-create_mergejoin_path(RelOptInfo *joinrel,
+MergePath *
+create_mergejoin_path(Query *root,
+					  RelOptInfo *joinrel,
+					  JoinType jointype,
 					  Path *outer_path,
 					  Path *inner_path,
 					  List *restrict_clauses,
@@ -465,6 +536,7 @@ create_mergejoin_path(RelOptInfo *joinrel,
 
 	pathnode->jpath.path.pathtype = T_MergeJoin;
 	pathnode->jpath.path.parent = joinrel;
+	pathnode->jpath.jointype = jointype;
 	pathnode->jpath.outerjoinpath = outer_path;
 	pathnode->jpath.innerjoinpath = inner_path;
 	pathnode->jpath.joinrestrictinfo = restrict_clauses;
@@ -474,9 +546,11 @@ create_mergejoin_path(RelOptInfo *joinrel,
 	pathnode->innersortkeys = innersortkeys;
 
 	cost_mergejoin(&pathnode->jpath.path,
+				   root,
 				   outer_path,
 				   inner_path,
 				   restrict_clauses,
+				   mergeclauses,
 				   outersortkeys,
 				   innersortkeys);
 
@@ -488,26 +562,27 @@ create_mergejoin_path(RelOptInfo *joinrel,
  *	  Creates a pathnode corresponding to a hash join between two relations.
  *
  * 'joinrel' is the join relation
+ * 'jointype' is the type of join required
  * 'outer_path' is the cheapest outer path
  * 'inner_path' is the cheapest inner path
  * 'restrict_clauses' are the RestrictInfo nodes to apply at the join
  * 'hashclauses' is a list of the hash join clause (always a 1-element list)
  *		(this should be a subset of the restrict_clauses list)
- * 'innerdisbursion' is an estimate of the disbursion of the inner hash key
- *
  */
-HashPath   *
-create_hashjoin_path(RelOptInfo *joinrel,
+HashPath *
+create_hashjoin_path(Query *root,
+					 RelOptInfo *joinrel,
+					 JoinType jointype,
 					 Path *outer_path,
 					 Path *inner_path,
 					 List *restrict_clauses,
-					 List *hashclauses,
-					 Selectivity innerdisbursion)
+					 List *hashclauses)
 {
 	HashPath   *pathnode = makeNode(HashPath);
 
 	pathnode->jpath.path.pathtype = T_HashJoin;
 	pathnode->jpath.path.parent = joinrel;
+	pathnode->jpath.jointype = jointype;
 	pathnode->jpath.outerjoinpath = outer_path;
 	pathnode->jpath.innerjoinpath = inner_path;
 	pathnode->jpath.joinrestrictinfo = restrict_clauses;
@@ -516,10 +591,11 @@ create_hashjoin_path(RelOptInfo *joinrel,
 	pathnode->path_hashclauses = hashclauses;
 
 	cost_hashjoin(&pathnode->jpath.path,
+				  root,
 				  outer_path,
 				  inner_path,
 				  restrict_clauses,
-				  innerdisbursion);
+				  hashclauses);
 
 	return pathnode;
 }
