@@ -2,23 +2,23 @@
  *
  * rewriteManip.c
  *
- * Portions Copyright (c) 1996-2000, PostgreSQL, Inc
+ * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/rewrite/rewriteManip.c,v 1.46 2000/04/12 17:15:32 momjian Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/rewrite/rewriteManip.c,v 1.61 2001/11/05 17:46:27 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
 #include "postgres.h"
 
+#include "nodes/makefuncs.h"
 #include "optimizer/clauses.h"
 #include "optimizer/tlist.h"
 #include "parser/parsetree.h"
 #include "parser/parse_clause.h"
 #include "rewrite/rewriteManip.h"
-#include "utils/builtins.h"
 #include "utils/lsyscache.h"
 
 
@@ -43,7 +43,15 @@ static bool checkExprHasSubLink_walker(Node *node, void *context);
 bool
 checkExprHasAggs(Node *node)
 {
-	return checkExprHasAggs_walker(node, NULL);
+	/*
+	 * If a Query is passed, examine it --- but we will not recurse into
+	 * sub-Queries.
+	 */
+	if (node && IsA(node, Query))
+		return query_tree_walker((Query *) node, checkExprHasAggs_walker,
+								 NULL, false);
+	else
+		return checkExprHasAggs_walker(node, NULL);
 }
 
 static bool
@@ -65,7 +73,15 @@ checkExprHasAggs_walker(Node *node, void *context)
 bool
 checkExprHasSubLink(Node *node)
 {
-	return checkExprHasSubLink_walker(node, NULL);
+	/*
+	 * If a Query is passed, examine it --- but we will not recurse into
+	 * sub-Queries.
+	 */
+	if (node && IsA(node, Query))
+		return query_tree_walker((Query *) node, checkExprHasSubLink_walker,
+								 NULL, false);
+	else
+		return checkExprHasSubLink_walker(node, NULL);
 }
 
 static bool
@@ -85,10 +101,11 @@ checkExprHasSubLink_walker(Node *node, void *context)
  *
  * Find all Var nodes in the given tree with varlevelsup == sublevels_up,
  * and increment their varno fields (rangetable indexes) by 'offset'.
- * The varnoold fields are adjusted similarly.
+ * The varnoold fields are adjusted similarly.	Also, RangeTblRef nodes
+ * in join trees and setOp trees are adjusted.
  *
  * NOTE: although this has the form of a walker, we cheat and modify the
- * Var nodes in-place.	The given expression tree should have been copied
+ * nodes in-place.	The given expression tree should have been copied
  * earlier to ensure that no unwanted side-effects occur!
  */
 
@@ -114,38 +131,25 @@ OffsetVarNodes_walker(Node *node, OffsetVarNodes_context *context)
 		}
 		return false;
 	}
-	if (IsA(node, SubLink))
+	if (IsA(node, RangeTblRef))
 	{
+		RangeTblRef *rtr = (RangeTblRef *) node;
 
-		/*
-		 * Standard expression_tree_walker will not recurse into
-		 * subselect, but here we must do so.
-		 */
-		SubLink    *sub = (SubLink *) node;
-
-		if (OffsetVarNodes_walker((Node *) (sub->lefthand),
-								  context))
-			return true;
-		OffsetVarNodes((Node *) (sub->subselect),
-					   context->offset,
-					   context->sublevels_up + 1);
+		if (context->sublevels_up == 0)
+			rtr->rtindex += context->offset;
+		/* the subquery itself is visited separately */
 		return false;
 	}
 	if (IsA(node, Query))
 	{
-		/* Reach here after recursing down into subselect above... */
-		Query	   *qry = (Query *) node;
+		/* Recurse into subselects */
+		bool		result;
 
-		if (OffsetVarNodes_walker((Node *) (qry->targetList),
-								  context))
-			return true;
-		if (OffsetVarNodes_walker((Node *) (qry->qual),
-								  context))
-			return true;
-		if (OffsetVarNodes_walker((Node *) (qry->havingQual),
-								  context))
-			return true;
-		return false;
+		context->sublevels_up++;
+		result = query_tree_walker((Query *) node, OffsetVarNodes_walker,
+								   (void *) context, true);
+		context->sublevels_up--;
+		return result;
 	}
 	return expression_tree_walker(node, OffsetVarNodes_walker,
 								  (void *) context);
@@ -158,7 +162,36 @@ OffsetVarNodes(Node *node, int offset, int sublevels_up)
 
 	context.offset = offset;
 	context.sublevels_up = sublevels_up;
-	OffsetVarNodes_walker(node, &context);
+
+	/*
+	 * Must be prepared to start with a Query or a bare expression tree;
+	 * if it's a Query, go straight to query_tree_walker to make sure that
+	 * sublevels_up doesn't get incremented prematurely.
+	 */
+	if (node && IsA(node, Query))
+	{
+		Query	   *qry = (Query *) node;
+		List	   *l;
+
+		/*
+		 * If we are starting at a Query, and sublevels_up is zero, then
+		 * we must also fix rangetable indexes in the Query itself ---
+		 * namely resultRelation and rowMarks entries.	sublevels_up
+		 * cannot be zero when recursing into a subquery, so there's no
+		 * need to have the same logic inside OffsetVarNodes_walker.
+		 */
+		if (sublevels_up == 0)
+		{
+			if (qry->resultRelation)
+				qry->resultRelation += offset;
+			foreach(l, qry->rowMarks)
+				lfirsti(l) += offset;
+		}
+		query_tree_walker(qry, OffsetVarNodes_walker,
+						  (void *) &context, true);
+	}
+	else
+		OffsetVarNodes_walker(node, &context);
 }
 
 /*
@@ -166,10 +199,11 @@ OffsetVarNodes(Node *node, int offset, int sublevels_up)
  *
  * Find all Var nodes in the given tree belonging to a specific relation
  * (identified by sublevels_up and rt_index), and change their varno fields
- * to 'new_index'.	The varnoold fields are changed too.
+ * to 'new_index'.	The varnoold fields are changed too.  Also, RangeTblRef
+ * nodes in join trees and setOp trees are adjusted.
  *
  * NOTE: although this has the form of a walker, we cheat and modify the
- * Var nodes in-place.	The given expression tree should have been copied
+ * nodes in-place.	The given expression tree should have been copied
  * earlier to ensure that no unwanted side-effects occur!
  */
 
@@ -197,39 +231,26 @@ ChangeVarNodes_walker(Node *node, ChangeVarNodes_context *context)
 		}
 		return false;
 	}
-	if (IsA(node, SubLink))
+	if (IsA(node, RangeTblRef))
 	{
+		RangeTblRef *rtr = (RangeTblRef *) node;
 
-		/*
-		 * Standard expression_tree_walker will not recurse into
-		 * subselect, but here we must do so.
-		 */
-		SubLink    *sub = (SubLink *) node;
-
-		if (ChangeVarNodes_walker((Node *) (sub->lefthand),
-								  context))
-			return true;
-		ChangeVarNodes((Node *) (sub->subselect),
-					   context->rt_index,
-					   context->new_index,
-					   context->sublevels_up + 1);
+		if (context->sublevels_up == 0 &&
+			rtr->rtindex == context->rt_index)
+			rtr->rtindex = context->new_index;
+		/* the subquery itself is visited separately */
 		return false;
 	}
 	if (IsA(node, Query))
 	{
-		/* Reach here after recursing down into subselect above... */
-		Query	   *qry = (Query *) node;
+		/* Recurse into subselects */
+		bool		result;
 
-		if (ChangeVarNodes_walker((Node *) (qry->targetList),
-								  context))
-			return true;
-		if (ChangeVarNodes_walker((Node *) (qry->qual),
-								  context))
-			return true;
-		if (ChangeVarNodes_walker((Node *) (qry->havingQual),
-								  context))
-			return true;
-		return false;
+		context->sublevels_up++;
+		result = query_tree_walker((Query *) node, ChangeVarNodes_walker,
+								   (void *) context, true);
+		context->sublevels_up--;
+		return result;
 	}
 	return expression_tree_walker(node, ChangeVarNodes_walker,
 								  (void *) context);
@@ -243,7 +264,39 @@ ChangeVarNodes(Node *node, int rt_index, int new_index, int sublevels_up)
 	context.rt_index = rt_index;
 	context.new_index = new_index;
 	context.sublevels_up = sublevels_up;
-	ChangeVarNodes_walker(node, &context);
+
+	/*
+	 * Must be prepared to start with a Query or a bare expression tree;
+	 * if it's a Query, go straight to query_tree_walker to make sure that
+	 * sublevels_up doesn't get incremented prematurely.
+	 */
+	if (node && IsA(node, Query))
+	{
+		Query	   *qry = (Query *) node;
+		List	   *l;
+
+		/*
+		 * If we are starting at a Query, and sublevels_up is zero, then
+		 * we must also fix rangetable indexes in the Query itself ---
+		 * namely resultRelation and rowMarks entries.	sublevels_up
+		 * cannot be zero when recursing into a subquery, so there's no
+		 * need to have the same logic inside ChangeVarNodes_walker.
+		 */
+		if (sublevels_up == 0)
+		{
+			if (qry->resultRelation == rt_index)
+				qry->resultRelation = new_index;
+			foreach(l, qry->rowMarks)
+			{
+				if (lfirsti(l) == rt_index)
+					lfirsti(l) = new_index;
+			}
+		}
+		query_tree_walker(qry, ChangeVarNodes_walker,
+						  (void *) &context, true);
+	}
+	else
+		ChangeVarNodes_walker(node, &context);
 }
 
 /*
@@ -283,38 +336,17 @@ IncrementVarSublevelsUp_walker(Node *node,
 			var->varlevelsup += context->delta_sublevels_up;
 		return false;
 	}
-	if (IsA(node, SubLink))
-	{
-
-		/*
-		 * Standard expression_tree_walker will not recurse into
-		 * subselect, but here we must do so.
-		 */
-		SubLink    *sub = (SubLink *) node;
-
-		if (IncrementVarSublevelsUp_walker((Node *) (sub->lefthand),
-										   context))
-			return true;
-		IncrementVarSublevelsUp((Node *) (sub->subselect),
-								context->delta_sublevels_up,
-								context->min_sublevels_up + 1);
-		return false;
-	}
 	if (IsA(node, Query))
 	{
-		/* Reach here after recursing down into subselect above... */
-		Query	   *qry = (Query *) node;
+		/* Recurse into subselects */
+		bool		result;
 
-		if (IncrementVarSublevelsUp_walker((Node *) (qry->targetList),
-										   context))
-			return true;
-		if (IncrementVarSublevelsUp_walker((Node *) (qry->qual),
-										   context))
-			return true;
-		if (IncrementVarSublevelsUp_walker((Node *) (qry->havingQual),
-										   context))
-			return true;
-		return false;
+		context->min_sublevels_up++;
+		result = query_tree_walker((Query *) node,
+								   IncrementVarSublevelsUp_walker,
+								   (void *) context, true);
+		context->min_sublevels_up--;
+		return result;
 	}
 	return expression_tree_walker(node, IncrementVarSublevelsUp_walker,
 								  (void *) context);
@@ -328,29 +360,260 @@ IncrementVarSublevelsUp(Node *node, int delta_sublevels_up,
 
 	context.delta_sublevels_up = delta_sublevels_up;
 	context.min_sublevels_up = min_sublevels_up;
-	IncrementVarSublevelsUp_walker(node, &context);
+
+	/*
+	 * Must be prepared to start with a Query or a bare expression tree;
+	 * if it's a Query, go straight to query_tree_walker to make sure that
+	 * sublevels_up doesn't get incremented prematurely.
+	 */
+	if (node && IsA(node, Query))
+		query_tree_walker((Query *) node, IncrementVarSublevelsUp_walker,
+						  (void *) &context, true);
+	else
+		IncrementVarSublevelsUp_walker(node, &context);
 }
 
+
+/*
+ * rangeTableEntry_used - detect whether an RTE is referenced somewhere
+ *	in var nodes or join or setOp trees of a query or expression.
+ */
+
+typedef struct
+{
+	int			rt_index;
+	int			sublevels_up;
+} rangeTableEntry_used_context;
+
+static bool
+rangeTableEntry_used_walker(Node *node,
+							rangeTableEntry_used_context *context)
+{
+	if (node == NULL)
+		return false;
+	if (IsA(node, Var))
+	{
+		Var		   *var = (Var *) node;
+
+		if (var->varlevelsup == context->sublevels_up &&
+			var->varno == context->rt_index)
+			return true;
+		return false;
+	}
+	if (IsA(node, RangeTblRef))
+	{
+		RangeTblRef *rtr = (RangeTblRef *) node;
+
+		if (rtr->rtindex == context->rt_index &&
+			context->sublevels_up == 0)
+			return true;
+		/* the subquery itself is visited separately */
+		return false;
+	}
+	if (IsA(node, Query))
+	{
+		/* Recurse into subselects */
+		bool		result;
+
+		context->sublevels_up++;
+		result = query_tree_walker((Query *) node, rangeTableEntry_used_walker,
+								   (void *) context, true);
+		context->sublevels_up--;
+		return result;
+	}
+	return expression_tree_walker(node, rangeTableEntry_used_walker,
+								  (void *) context);
+}
+
+bool
+rangeTableEntry_used(Node *node, int rt_index, int sublevels_up)
+{
+	rangeTableEntry_used_context context;
+
+	context.rt_index = rt_index;
+	context.sublevels_up = sublevels_up;
+
+	/*
+	 * Must be prepared to start with a Query or a bare expression tree;
+	 * if it's a Query, go straight to query_tree_walker to make sure that
+	 * sublevels_up doesn't get incremented prematurely.
+	 */
+	if (node && IsA(node, Query))
+		return query_tree_walker((Query *) node, rangeTableEntry_used_walker,
+								 (void *) &context, true);
+	else
+		return rangeTableEntry_used_walker(node, &context);
+}
+
+
+/*
+ * attribute_used -
+ *	Check if a specific attribute number of a RTE is used
+ *	somewhere in the query or expression.
+ */
+
+typedef struct
+{
+	int			rt_index;
+	int			attno;
+	int			sublevels_up;
+} attribute_used_context;
+
+static bool
+attribute_used_walker(Node *node,
+					  attribute_used_context *context)
+{
+	if (node == NULL)
+		return false;
+	if (IsA(node, Var))
+	{
+		Var		   *var = (Var *) node;
+
+		if (var->varlevelsup == context->sublevels_up &&
+			var->varno == context->rt_index &&
+			var->varattno == context->attno)
+			return true;
+		return false;
+	}
+	if (IsA(node, Query))
+	{
+		/* Recurse into subselects */
+		bool		result;
+
+		context->sublevels_up++;
+		result = query_tree_walker((Query *) node, attribute_used_walker,
+								   (void *) context, true);
+		context->sublevels_up--;
+		return result;
+	}
+	return expression_tree_walker(node, attribute_used_walker,
+								  (void *) context);
+}
+
+bool
+attribute_used(Node *node, int rt_index, int attno, int sublevels_up)
+{
+	attribute_used_context context;
+
+	context.rt_index = rt_index;
+	context.attno = attno;
+	context.sublevels_up = sublevels_up;
+
+	/*
+	 * Must be prepared to start with a Query or a bare expression tree;
+	 * if it's a Query, go straight to query_tree_walker to make sure that
+	 * sublevels_up doesn't get incremented prematurely.
+	 */
+	if (node && IsA(node, Query))
+		return query_tree_walker((Query *) node, attribute_used_walker,
+								 (void *) &context, true);
+	else
+		return attribute_used_walker(node, &context);
+}
+
+
+/*
+ * If the given Query is an INSERT ... SELECT construct, extract and
+ * return the sub-Query node that represents the SELECT part.  Otherwise
+ * return the given Query.
+ *
+ * If subquery_ptr is not NULL, then *subquery_ptr is set to the location
+ * of the link to the SELECT subquery inside parsetree, or NULL if not an
+ * INSERT ... SELECT.
+ *
+ * This is a hack needed because transformations on INSERT ... SELECTs that
+ * appear in rule actions should be applied to the source SELECT, not to the
+ * INSERT part.  Perhaps this can be cleaned up with redesigned querytrees.
+ */
+Query *
+getInsertSelectQuery(Query *parsetree, Query ***subquery_ptr)
+{
+	Query	   *selectquery;
+	RangeTblEntry *selectrte;
+	RangeTblRef *rtr;
+
+	if (subquery_ptr)
+		*subquery_ptr = NULL;
+
+	if (parsetree == NULL)
+		return parsetree;
+	if (parsetree->commandType != CMD_INSERT)
+		return parsetree;
+
+	/*
+	 * Currently, this is ONLY applied to rule-action queries, and so we
+	 * expect to find the *OLD* and *NEW* placeholder entries in the given
+	 * query.  If they're not there, it must be an INSERT/SELECT in which
+	 * they've been pushed down to the SELECT.
+	 */
+	if (length(parsetree->rtable) >= 2 &&
+		strcmp(rt_fetch(PRS2_OLD_VARNO, parsetree->rtable)->eref->relname,
+			   "*OLD*") == 0 &&
+		strcmp(rt_fetch(PRS2_NEW_VARNO, parsetree->rtable)->eref->relname,
+			   "*NEW*") == 0)
+		return parsetree;
+	Assert(parsetree->jointree && IsA(parsetree->jointree, FromExpr));
+	if (length(parsetree->jointree->fromlist) != 1)
+		elog(ERROR, "getInsertSelectQuery: expected to find SELECT subquery");
+	rtr = (RangeTblRef *) lfirst(parsetree->jointree->fromlist);
+	Assert(IsA(rtr, RangeTblRef));
+	selectrte = rt_fetch(rtr->rtindex, parsetree->rtable);
+	selectquery = selectrte->subquery;
+	if (!(selectquery && IsA(selectquery, Query) &&
+		  selectquery->commandType == CMD_SELECT))
+		elog(ERROR, "getInsertSelectQuery: expected to find SELECT subquery");
+	if (length(selectquery->rtable) >= 2 &&
+	 strcmp(rt_fetch(PRS2_OLD_VARNO, selectquery->rtable)->eref->relname,
+			"*OLD*") == 0 &&
+	 strcmp(rt_fetch(PRS2_NEW_VARNO, selectquery->rtable)->eref->relname,
+			"*NEW*") == 0)
+	{
+		if (subquery_ptr)
+			*subquery_ptr = &(selectrte->subquery);
+		return selectquery;
+	}
+	elog(ERROR, "getInsertSelectQuery: can't find rule placeholders");
+	return NULL;				/* not reached */
+}
+
+
 /*
  * Add the given qualifier condition to the query's WHERE clause
  */
 void
 AddQual(Query *parsetree, Node *qual)
 {
-	Node	   *copy,
-			   *old;
+	Node	   *copy;
 
 	if (qual == NULL)
 		return;
 
+	if (parsetree->commandType == CMD_UTILITY)
+	{
+		/*
+		 * There's noplace to put the qual on a utility statement.
+		 *
+		 * If it's a NOTIFY, silently ignore the qual; this means that the
+		 * NOTIFY will execute, whether or not there are any qualifying
+		 * rows. While clearly wrong, this is much more useful than
+		 * refusing to execute the rule at all, and extra NOTIFY events
+		 * are harmless for typical uses of NOTIFY.
+		 *
+		 * If it isn't a NOTIFY, error out, since unconditional execution of
+		 * other utility stmts is unlikely to be wanted.  (This case is
+		 * not currently allowed anyway, but keep the test for safety.)
+		 */
+		if (parsetree->utilityStmt && IsA(parsetree->utilityStmt, NotifyStmt))
+			return;
+		else
+			elog(ERROR, "Conditional utility statements are not implemented");
+	}
+
 	/* INTERSECT want's the original, but we need to copy - Jan */
 	copy = copyObject(qual);
 
-	old = parsetree->qual;
-	if (old == NULL)
-		parsetree->qual = copy;
-	else
-		parsetree->qual = (Node *) make_andclause(makeList(old, copy, -1));
+	parsetree->jointree->quals = make_and_qual(parsetree->jointree->quals,
+											   copy);
 
 	/*
 	 * Make sure query is marked correctly if added qual has sublinks or
@@ -368,20 +631,29 @@ AddQual(Query *parsetree, Node *qual)
 void
 AddHavingQual(Query *parsetree, Node *havingQual)
 {
-	Node	   *copy,
-			   *old;
+	Node	   *copy;
 
 	if (havingQual == NULL)
 		return;
 
+	if (parsetree->commandType == CMD_UTILITY)
+	{
+		/*
+		 * There's noplace to put the qual on a utility statement.
+		 *
+		 * See comments in AddQual for motivation.
+		 */
+		if (parsetree->utilityStmt && IsA(parsetree->utilityStmt, NotifyStmt))
+			return;
+		else
+			elog(ERROR, "Conditional utility statements are not implemented");
+	}
+
 	/* INTERSECT want's the original, but we need to copy - Jan */
 	copy = copyObject(havingQual);
 
-	old = parsetree->havingQual;
-	if (old == NULL)
-		parsetree->havingQual = copy;
-	else
-		parsetree->havingQual = (Node *) make_andclause(makeList(old, copy, -1));
+	parsetree->havingQual = make_and_qual(parsetree->havingQual,
+										  copy);
 
 	/*
 	 * Make sure query is marked correctly if added qual has sublinks or
@@ -406,7 +678,6 @@ AddNotHavingQual(Query *parsetree, Node *havingQual)
 
 	AddHavingQual(parsetree, notqual);
 }
-
 #endif
 
 void
@@ -424,80 +695,6 @@ AddNotQual(Query *parsetree, Node *qual)
 }
 
 
-/*
- * Add all expressions used by the given GroupClause list to the
- * parsetree's targetlist and groupclause list.
- *
- * tlist is the old targetlist associated with the input groupclauses.
- *
- * XXX shouldn't we be checking to see if there are already matching
- * entries in parsetree->targetlist?
- */
-void
-AddGroupClause(Query *parsetree, List *group_by, List *tlist)
-{
-	List	   *l;
-
-	foreach(l, group_by)
-	{
-		GroupClause *groupclause = (GroupClause *) copyObject(lfirst(l));
-		TargetEntry *tle = get_sortgroupclause_tle(groupclause, tlist);
-
-		/* copy the groupclause's TLE from the old tlist */
-		tle = (TargetEntry *) copyObject(tle);
-
-		/*
-		 * The ressortgroupref number in the old tlist might be already
-		 * taken in the new tlist, so force assignment of a new number.
-		 */
-		tle->resdom->ressortgroupref = 0;
-		groupclause->tleSortGroupRef =
-			assignSortGroupRef(tle, parsetree->targetList);
-
-		/* Also need to set the resno and mark it resjunk. */
-		tle->resdom->resno = length(parsetree->targetList) + 1;
-		tle->resdom->resjunk = true;
-
-		parsetree->targetList = lappend(parsetree->targetList, tle);
-		parsetree->groupClause = lappend(parsetree->groupClause, groupclause);
-	}
-}
-
-static Node *
-make_null(Oid type)
-{
-	Const	   *c = makeNode(Const);
-
-	c->consttype = type;
-	c->constlen = get_typlen(type);
-	c->constvalue = PointerGetDatum(NULL);
-	c->constisnull = true;
-	c->constbyval = get_typbyval(type);
-	return (Node *) c;
-}
-
-#ifdef NOT_USED
-void
-FixResdomTypes(List *tlist)
-{
-	List	   *i;
-
-	foreach(i, tlist)
-	{
-		TargetEntry *tle = lfirst(i);
-
-		if (nodeTag(tle->expr) == T_Var)
-		{
-			Var		   *var = (Var *) tle->expr;
-
-			tle->resdom->restype = var->vartype;
-			tle->resdom->restypmod = var->vartypmod;
-		}
-	}
-}
-
-#endif
-
 /* Find a targetlist entry by resno */
 static Node *
 FindMatchingNew(List *tlist, int attno)
@@ -514,6 +711,8 @@ FindMatchingNew(List *tlist, int attno)
 	return NULL;
 }
 
+#ifdef NOT_USED
+
 /* Find a targetlist entry by resname */
 static Node *
 FindMatchingTLEntry(List *tlist, char *e_attname)
@@ -526,25 +725,30 @@ FindMatchingTLEntry(List *tlist, char *e_attname)
 		char	   *resname;
 
 		resname = tle->resdom->resname;
-		if (!strcmp(e_attname, resname))
+		if (strcmp(e_attname, resname) == 0)
 			return tle->expr;
 	}
 	return NULL;
 }
+#endif
 
 
 /*
  * ResolveNew - replace Vars with corresponding items from a targetlist
  *
- * Vars matching info->new_varno and sublevels_up are replaced by the
+ * Vars matching target_varno and sublevels_up are replaced by the
  * entry with matching resno from targetlist, if there is one.
+ * If not, we either change the unmatched Var's varno to update_varno
+ * (when event == CMD_UPDATE) or replace it with a constant NULL.
  */
 
 typedef struct
 {
-	RewriteInfo *info;
-	List	   *targetlist;
+	int			target_varno;
 	int			sublevels_up;
+	List	   *targetlist;
+	int			event;
+	int			update_varno;
 } ResolveNew_context;
 
 static Node *
@@ -558,26 +762,31 @@ ResolveNew_mutator(Node *node, ResolveNew_context *context)
 		int			this_varno = (int) var->varno;
 		int			this_varlevelsup = (int) var->varlevelsup;
 
-		if (this_varno == context->info->new_varno &&
+		if (this_varno == context->target_varno &&
 			this_varlevelsup == context->sublevels_up)
 		{
-			Node	   *n = FindMatchingNew(context->targetlist,
-											var->varattno);
+			Node	   *n;
+
+			/* band-aid: don't do the wrong thing with a whole-tuple Var */
+			if (var->varattno == InvalidAttrNumber)
+				elog(ERROR, "ResolveNew: can't handle whole-tuple reference");
+
+			n = FindMatchingNew(context->targetlist, var->varattno);
 
 			if (n == NULL)
 			{
-				if (context->info->event == CMD_UPDATE)
+				if (context->event == CMD_UPDATE)
 				{
 					/* For update, just change unmatched var's varno */
-					n = copyObject(node);
-					((Var *) n)->varno = context->info->current_varno;
-					((Var *) n)->varnoold = context->info->current_varno;
-					return n;
+					var = (Var *) copyObject(node);
+					var->varno = context->update_varno;
+					var->varnoold = context->update_varno;
+					return (Node *) var;
 				}
 				else
 				{
 					/* Otherwise replace unmatched var with a null */
-					return make_null(var->vartype);
+					return (Node *) makeNullConst(var->vartype);
 				}
 			}
 			else
@@ -605,10 +814,8 @@ ResolveNew_mutator(Node *node, ResolveNew_context *context)
 		FLATCOPY(newnode, sublink, SubLink);
 		MUTATE(newnode->lefthand, sublink->lefthand, List *,
 			   ResolveNew_mutator, context);
-		context->sublevels_up++;
 		MUTATE(newnode->subselect, sublink->subselect, Node *,
 			   ResolveNew_mutator, context);
-		context->sublevels_up--;
 		return (Node *) newnode;
 	}
 	if (IsA(node, Query))
@@ -616,49 +823,49 @@ ResolveNew_mutator(Node *node, ResolveNew_context *context)
 		Query	   *query = (Query *) node;
 		Query	   *newnode;
 
-		/*
-		 * XXX original code for ResolveNew only recursed into qual field
-		 * of subquery.  I'm assuming that was an oversight ... tgl 9/99
-		 */
-
 		FLATCOPY(newnode, query, Query);
-		MUTATE(newnode->targetList, query->targetList, List *,
-			   ResolveNew_mutator, context);
-		MUTATE(newnode->qual, query->qual, Node *,
-			   ResolveNew_mutator, context);
-		MUTATE(newnode->havingQual, query->havingQual, Node *,
-			   ResolveNew_mutator, context);
+		context->sublevels_up++;
+		query_tree_mutator(newnode, ResolveNew_mutator, context, true);
+		context->sublevels_up--;
 		return (Node *) newnode;
 	}
 	return expression_tree_mutator(node, ResolveNew_mutator,
 								   (void *) context);
 }
 
-static Node *
-ResolveNew(Node *node, RewriteInfo *info, List *targetlist,
-		   int sublevels_up)
+Node *
+ResolveNew(Node *node, int target_varno, int sublevels_up,
+		   List *targetlist, int event, int update_varno)
 {
 	ResolveNew_context context;
 
-	context.info = info;
-	context.targetlist = targetlist;
+	context.target_varno = target_varno;
 	context.sublevels_up = sublevels_up;
+	context.targetlist = targetlist;
+	context.event = event;
+	context.update_varno = update_varno;
 
-	return ResolveNew_mutator(node, &context);
+	/*
+	 * Must be prepared to start with a Query or a bare expression tree;
+	 * if it's a Query, go straight to query_tree_mutator to make sure
+	 * that sublevels_up doesn't get incremented prematurely.
+	 */
+	if (node && IsA(node, Query))
+	{
+		Query	   *query = (Query *) node;
+		Query	   *newnode;
+
+		FLATCOPY(newnode, query, Query);
+		query_tree_mutator(newnode, ResolveNew_mutator,
+						   (void *) &context, true);
+		return (Node *) newnode;
+	}
+	else
+		return ResolveNew_mutator(node, &context);
 }
 
-void
-FixNew(RewriteInfo *info, Query *parsetree)
-{
-	info->rule_action->targetList = (List *)
-	ResolveNew((Node *) info->rule_action->targetList,
-			   info, parsetree->targetList, 0);
-	info->rule_action->qual = ResolveNew(info->rule_action->qual,
-										 info, parsetree->targetList, 0);
-	/* XXX original code didn't fix havingQual; presumably an oversight? */
-	info->rule_action->havingQual = ResolveNew(info->rule_action->havingQual,
-										 info, parsetree->targetList, 0);
-}
+
+#ifdef NOT_USED
 
 /*
  * HandleRIRAttributeRule
@@ -666,8 +873,6 @@ FixNew(RewriteInfo *info, Query *parsetree)
  *
  * Handles 'on retrieve to relation.attribute
  *			do instead retrieve (attribute = expression) w/qual'
- *
- * XXX Why is this not unified with apply_RIR_view()?
  */
 
 typedef struct
@@ -679,11 +884,11 @@ typedef struct
 	int		   *modified;
 	int		   *badsql;
 	int			sublevels_up;
-} HandleRIRAttributeRule_context;
+}	HandleRIRAttributeRule_context;
 
 static Node *
 HandleRIRAttributeRule_mutator(Node *node,
-							   HandleRIRAttributeRule_context *context)
+							   HandleRIRAttributeRule_context * context)
 {
 	if (node == NULL)
 		return NULL;
@@ -702,7 +907,7 @@ HandleRIRAttributeRule_mutator(Node *node,
 			{					/* HACK: disallow SET variables */
 				*context->modified = TRUE;
 				*context->badsql = TRUE;
-				return make_null(var->vartype);
+				return (Node *) makeNullConst(var->vartype);
 			}
 			else
 			{
@@ -719,7 +924,7 @@ HandleRIRAttributeRule_mutator(Node *node,
 					n = FindMatchingTLEntry(context->targetlist,
 											name_to_look_for);
 					if (n == NULL)
-						return make_null(var->vartype);
+						return (Node *) makeNullConst(var->vartype);
 					/* Make a copy of the tlist item to return */
 					n = copyObject(n);
 
@@ -748,10 +953,8 @@ HandleRIRAttributeRule_mutator(Node *node,
 		FLATCOPY(newnode, sublink, SubLink);
 		MUTATE(newnode->lefthand, sublink->lefthand, List *,
 			   HandleRIRAttributeRule_mutator, context);
-		context->sublevels_up++;
 		MUTATE(newnode->subselect, sublink->subselect, Node *,
 			   HandleRIRAttributeRule_mutator, context);
-		context->sublevels_up--;
 		return (Node *) newnode;
 	}
 	if (IsA(node, Query))
@@ -759,18 +962,11 @@ HandleRIRAttributeRule_mutator(Node *node,
 		Query	   *query = (Query *) node;
 		Query	   *newnode;
 
-		/*
-		 * XXX original code for HandleRIRAttributeRule only recursed into
-		 * qual field of subquery.	I'm assuming that was an oversight ...
-		 */
-
 		FLATCOPY(newnode, query, Query);
-		MUTATE(newnode->targetList, query->targetList, List *,
-			   HandleRIRAttributeRule_mutator, context);
-		MUTATE(newnode->qual, query->qual, Node *,
-			   HandleRIRAttributeRule_mutator, context);
-		MUTATE(newnode->havingQual, query->havingQual, Node *,
-			   HandleRIRAttributeRule_mutator, context);
+		context->sublevels_up++;
+		query_tree_mutator(newnode, HandleRIRAttributeRule_mutator,
+						   context, true);
+		context->sublevels_up--;
 		return (Node *) newnode;
 	}
 	return expression_tree_mutator(node, HandleRIRAttributeRule_mutator,
@@ -796,12 +992,8 @@ HandleRIRAttributeRule(Query *parsetree,
 	context.badsql = badsql;
 	context.sublevels_up = 0;
 
-	parsetree->targetList = (List *)
-		HandleRIRAttributeRule_mutator((Node *) parsetree->targetList,
-									   &context);
-	parsetree->qual = HandleRIRAttributeRule_mutator(parsetree->qual,
-													 &context);
-	/* XXX original code did not fix havingQual ... oversight? */
-	parsetree->havingQual = HandleRIRAttributeRule_mutator(parsetree->havingQual,
-														   &context);
+	query_tree_mutator(parsetree, HandleRIRAttributeRule_mutator,
+					   (void *) &context, true);
 }
+
+#endif   /* NOT_USED */
