@@ -8,7 +8,7 @@
  *
  *
  * IDENTIFICATION
- *	  $Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/storage/ipc/sinval.c,v 1.29 2001/03/23 04:49:54 momjian Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/storage/ipc/sinval.c,v 1.43 2001/10/25 05:49:42 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -16,13 +16,12 @@
 
 #include <sys/types.h>
 
-#include "storage/backendid.h"
 #include "storage/proc.h"
 #include "storage/sinval.h"
 #include "storage/sinvaladt.h"
 #include "utils/tqual.h"
+#include "miscadmin.h"
 
-SPINLOCK	SInvalLock = (SPINLOCK) NULL;
 
 /****************************************************************************/
 /*	CreateSharedInvalidationState()		 Initialize SI buffer				*/
@@ -32,7 +31,7 @@ SPINLOCK	SInvalLock = (SPINLOCK) NULL;
 void
 CreateSharedInvalidationState(int maxBackends)
 {
-	/* SInvalLock must be initialized already, during spinlock init */
+	/* SInvalLock must be initialized already, during LWLock init */
 	SIBufferInit(maxBackends);
 }
 
@@ -43,88 +42,79 @@ CreateSharedInvalidationState(int maxBackends)
 void
 InitBackendSharedInvalidationState(void)
 {
-	SpinAcquire(SInvalLock);
-	if (!SIBackendInit(shmInvalBuffer))
-	{
-		SpinRelease(SInvalLock);
+	int			flag;
+
+	LWLockAcquire(SInvalLock, LW_EXCLUSIVE);
+	flag = SIBackendInit(shmInvalBuffer);
+	LWLockRelease(SInvalLock);
+	if (flag < 0)				/* unexpected problem */
 		elog(FATAL, "Backend cache invalidation initialization failed");
-	}
-	SpinRelease(SInvalLock);
+	if (flag == 0)				/* expected problem: MaxBackends exceeded */
+		elog(FATAL, "Sorry, too many clients already");
 }
 
 /*
- * RegisterSharedInvalid
+ * SendSharedInvalidMessage
  *	Add a shared-cache-invalidation message to the global SI message queue.
- *
- * Note:
- *	Assumes hash index is valid.
- *	Assumes item pointer is valid.
  */
 void
-RegisterSharedInvalid(int cacheId,		/* XXX */
-					  Index hashIndex,
-					  ItemPointer pointer)
+SendSharedInvalidMessage(SharedInvalidationMessage *msg)
 {
-	SharedInvalidData newInvalid;
 	bool		insertOK;
 
-	/*
-	 * This code has been hacked to accept two types of messages.  This
-	 * might be treated more generally in the future.
-	 *
-	 * (1) cacheId= system cache id hashIndex= system cache hash index for a
-	 * (possibly) cached tuple pointer= pointer of (possibly) cached tuple
-	 *
-	 * (2) cacheId= special non-syscache id hashIndex= object id contained in
-	 * (possibly) cached relation descriptor pointer= null
-	 */
-
-	newInvalid.cacheId = cacheId;
-	newInvalid.hashIndex = hashIndex;
-
-	if (ItemPointerIsValid(pointer))
-		ItemPointerCopy(pointer, &newInvalid.pointerData);
-	else
-		ItemPointerSetInvalid(&newInvalid.pointerData);
-
-	SpinAcquire(SInvalLock);
-	insertOK = SIInsertDataEntry(shmInvalBuffer, &newInvalid);
-	SpinRelease(SInvalLock);
+	LWLockAcquire(SInvalLock, LW_EXCLUSIVE);
+	insertOK = SIInsertDataEntry(shmInvalBuffer, msg);
+	LWLockRelease(SInvalLock);
 	if (!insertOK)
-		elog(NOTICE, "RegisterSharedInvalid: SI buffer overflow");
+		elog(DEBUG, "SendSharedInvalidMessage: SI buffer overflow");
 }
 
 /*
- * InvalidateSharedInvalid
+ * ReceiveSharedInvalidMessages
  *		Process shared-cache-invalidation messages waiting for this backend
  */
 void
-			InvalidateSharedInvalid(void (*invalFunction) (),
-									void (*resetFunction) ())
+ReceiveSharedInvalidMessages(
+				  void (*invalFunction) (SharedInvalidationMessage *msg),
+							 void (*resetFunction) (void))
 {
-	SharedInvalidData data;
+	SharedInvalidationMessage data;
 	int			getResult;
 	bool		gotMessage = false;
 
 	for (;;)
 	{
-		SpinAcquire(SInvalLock);
+		/*
+		 * We can run SIGetDataEntry in parallel with other backends
+		 * running SIGetDataEntry for themselves, since each instance will
+		 * modify only fields of its own backend's ProcState, and no
+		 * instance will look at fields of other backends' ProcStates.  We
+		 * express this by grabbing SInvalLock in shared mode.	Note that
+		 * this is not exactly the normal (read-only) interpretation of a
+		 * shared lock! Look closely at the interactions before allowing
+		 * SInvalLock to be grabbed in shared mode for any other reason!
+		 *
+		 * The routines later in this file that use shared mode are okay with
+		 * this, because they aren't looking at the ProcState fields
+		 * associated with SI message transfer; they only use the
+		 * ProcState array as an easy way to find all the PROC structures.
+		 */
+		LWLockAcquire(SInvalLock, LW_SHARED);
 		getResult = SIGetDataEntry(shmInvalBuffer, MyBackendId, &data);
-		SpinRelease(SInvalLock);
+		LWLockRelease(SInvalLock);
+
 		if (getResult == 0)
 			break;				/* nothing more to do */
 		if (getResult < 0)
 		{
 			/* got a reset message */
-			elog(NOTICE, "InvalidateSharedInvalid: cache state reset");
+			elog(DEBUG, "ReceiveSharedInvalidMessages: cache state reset");
 			resetFunction();
 		}
 		else
 		{
 			/* got a normal data message */
-			invalFunction(data.cacheId,
-						  data.hashIndex,
-						  &data.pointerData);
+			invalFunction(&data);
 		}
 		gotMessage = true;
 	}
@@ -132,9 +122,9 @@ void
 	/* If we got any messages, try to release dead messages */
 	if (gotMessage)
 	{
-		SpinAcquire(SInvalLock);
+		LWLockAcquire(SInvalLock, LW_EXCLUSIVE);
 		SIDelExpiredDataEntries(shmInvalBuffer);
-		SpinRelease(SInvalLock);
+		LWLockRelease(SInvalLock);
 	}
 }
 
@@ -173,7 +163,7 @@ DatabaseHasActiveBackends(Oid databaseId, bool ignoreMyself)
 	ProcState  *stateP = segP->procState;
 	int			index;
 
-	SpinAcquire(SInvalLock);
+	LWLockAcquire(SInvalLock, LW_SHARED);
 
 	for (index = 0; index < segP->lastBackend; index++)
 	{
@@ -194,7 +184,7 @@ DatabaseHasActiveBackends(Oid databaseId, bool ignoreMyself)
 		}
 	}
 
-	SpinRelease(SInvalLock);
+	LWLockRelease(SInvalLock);
 
 	return result;
 }
@@ -210,7 +200,7 @@ TransactionIdIsInProgress(TransactionId xid)
 	ProcState  *stateP = segP->procState;
 	int			index;
 
-	SpinAcquire(SInvalLock);
+	LWLockAcquire(SInvalLock, LW_SHARED);
 
 	for (index = 0; index < segP->lastBackend; index++)
 	{
@@ -220,7 +210,10 @@ TransactionIdIsInProgress(TransactionId xid)
 		{
 			PROC	   *proc = (PROC *) MAKE_PTR(pOffset);
 
-			if (proc->xid == xid)
+			/* Fetch xid just once - see GetNewTransactionId */
+			TransactionId pxid = proc->xid;
+
+			if (TransactionIdEquals(pxid, xid))
 			{
 				result = true;
 				break;
@@ -228,27 +221,38 @@ TransactionIdIsInProgress(TransactionId xid)
 		}
 	}
 
-	SpinRelease(SInvalLock);
+	LWLockRelease(SInvalLock);
 
 	return result;
 }
 
 /*
- * GetXmaxRecent -- returns oldest transaction that was running
- *					when all current transaction was started.
- *					It's used by vacuum to decide what deleted
- *					tuples must be preserved in a table.
+ * GetOldestXmin -- returns oldest transaction that was running
+ *					when any current transaction was started.
+ *
+ * If allDbs is TRUE then all backends are considered; if allDbs is FALSE
+ * then only backends running in my own database are considered.
+ *
+ * This is used by VACUUM to decide which deleted tuples must be preserved
+ * in a table.	allDbs = TRUE is needed for shared relations, but allDbs =
+ * FALSE is sufficient for non-shared relations, since only backends in my
+ * own database could ever see the tuples in them.
+ *
+ * Note: we include the currently running xids in the set of considered xids.
+ * This ensures that if a just-started xact has not yet set its snapshot,
+ * when it does set the snapshot it cannot set xmin less than what we compute.
  */
-void
-GetXmaxRecent(TransactionId *XmaxRecent)
+TransactionId
+GetOldestXmin(bool allDbs)
 {
 	SISeg	   *segP = shmInvalBuffer;
 	ProcState  *stateP = segP->procState;
+	TransactionId result;
 	int			index;
 
-	*XmaxRecent = GetCurrentTransactionId();
+	result = GetCurrentTransactionId();
 
-	SpinAcquire(SInvalLock);
+	LWLockAcquire(SInvalLock, LW_SHARED);
 
 	for (index = 0; index < segP->lastBackend; index++)
 	{
@@ -257,22 +261,43 @@ GetXmaxRecent(TransactionId *XmaxRecent)
 		if (pOffset != INVALID_OFFSET)
 		{
 			PROC	   *proc = (PROC *) MAKE_PTR(pOffset);
-			TransactionId xmin;
 
-			xmin = proc->xmin;	/* we don't use spin-locking in
-								 * AbortTransaction() ! */
-			if (proc == MyProc || xmin < FirstTransactionId)
-				continue;
-			if (xmin < *XmaxRecent)
-				*XmaxRecent = xmin;
+			if (allDbs || proc->databaseId == MyDatabaseId)
+			{
+				/* Fetch xid just once - see GetNewTransactionId */
+				TransactionId xid = proc->xid;
+
+				if (TransactionIdIsNormal(xid))
+				{
+					if (TransactionIdPrecedes(xid, result))
+						result = xid;
+					xid = proc->xmin;
+					if (TransactionIdIsNormal(xid))
+						if (TransactionIdPrecedes(xid, result))
+							result = xid;
+				}
+			}
 		}
 	}
 
-	SpinRelease(SInvalLock);
+	LWLockRelease(SInvalLock);
+
+	return result;
 }
 
-/*
+/*----------
  * GetSnapshotData -- returns information about running transactions.
+ *
+ * The returned snapshot includes xmin (lowest still-running xact ID),
+ * xmax (next xact ID to be assigned), and a list of running xact IDs
+ * in the range xmin <= xid < xmax.  It is used as follows:
+ *		All xact IDs < xmin are considered finished.
+ *		All xact IDs >= xmax are considered still running.
+ *		For an xact ID xmin <= xid < xmax, consult list to see whether
+ *		it is considered running or not.
+ * This ensures that the set of transactions seen as "running" by the
+ * current xact will not change after it takes the snapshot.
+ *----------
  */
 Snapshot
 GetSnapshotData(bool serializable)
@@ -283,23 +308,55 @@ GetSnapshotData(bool serializable)
 	int			index;
 	int			count = 0;
 
+	if (snapshot == NULL)
+		elog(ERROR, "Memory exhausted in GetSnapshotData");
+
 	/*
-	 * There can be no more than lastBackend active transactions, so this
-	 * is enough space:
+	 * Allocating space for MaxBackends xids is usually overkill;
+	 * lastBackend would be sufficient.  But it seems better to do the
+	 * malloc while not holding the lock, so we can't look at lastBackend.
 	 */
 	snapshot->xip = (TransactionId *)
-		malloc(segP->lastBackend * sizeof(TransactionId));
+		malloc(MaxBackends * sizeof(TransactionId));
+	if (snapshot->xip == NULL)
+		elog(ERROR, "Memory exhausted in GetSnapshotData");
+
 	snapshot->xmin = GetCurrentTransactionId();
 
-	SpinAcquire(SInvalLock);
-
 	/*
-	 * Unfortunately, we have to call ReadNewTransactionId() after
-	 * acquiring SInvalLock above. It's not good because
-	 * ReadNewTransactionId() does SpinAcquire(XidGenLockId) but
-	 * _necessary_.
+	 * If we are going to set MyProc->xmin then we'd better get exclusive
+	 * lock; if not, this is a read-only operation so it can be shared.
 	 */
-	ReadNewTransactionId(&(snapshot->xmax));
+	LWLockAcquire(SInvalLock, serializable ? LW_EXCLUSIVE : LW_SHARED);
+
+	/*--------------------
+	 * Unfortunately, we have to call ReadNewTransactionId() after acquiring
+	 * SInvalLock above.  It's not good because ReadNewTransactionId() does
+	 * LWLockAcquire(XidGenLock), but *necessary*.	We need to be sure that
+	 * no transactions exit the set of currently-running transactions
+	 * between the time we fetch xmax and the time we finish building our
+	 * snapshot.  Otherwise we could have a situation like this:
+	 *
+	 *		1. Tx Old is running (in Read Committed mode).
+	 *		2. Tx S reads new transaction ID into xmax, then
+	 *		   is swapped out before acquiring SInvalLock.
+	 *		3. Tx New gets new transaction ID (>= S' xmax),
+	 *		   makes changes and commits.
+	 *		4. Tx Old changes some row R changed by Tx New and commits.
+	 *		5. Tx S finishes getting its snapshot data.  It sees Tx Old as
+	 *		   done, but sees Tx New as still running (since New >= xmax).
+	 *
+	 * Now S will see R changed by both Tx Old and Tx New, *but* does not
+	 * see other changes made by Tx New.  If S is supposed to be in
+	 * Serializable mode, this is wrong.
+	 *
+	 * By locking SInvalLock before we read xmax, we ensure that TX Old
+	 * cannot exit the set of running transactions seen by Tx S.  Therefore
+	 * both Old and New will be seen as still running => no inconsistency.
+	 *--------------------
+	 */
+
+	snapshot->xmax = ReadNewTransactionId();
 
 	for (index = 0; index < segP->lastBackend; index++)
 	{
@@ -308,28 +365,23 @@ GetSnapshotData(bool serializable)
 		if (pOffset != INVALID_OFFSET)
 		{
 			PROC	   *proc = (PROC *) MAKE_PTR(pOffset);
-			TransactionId xid;
+
+			/* Fetch xid just once - see GetNewTransactionId */
+			TransactionId xid = proc->xid;
 
 			/*
-			 * We don't use spin-locking when changing proc->xid in
-			 * GetNewTransactionId() and in AbortTransaction() !..
+			 * Ignore my own proc (dealt with my xid above), procs not
+			 * running a transaction, and xacts started since we read the
+			 * next transaction ID.  There's no need to store XIDs above
+			 * what we got from ReadNewTransactionId, since we'll treat
+			 * them as running anyway.
 			 */
-			xid = proc->xid;
 			if (proc == MyProc ||
-				xid < FirstTransactionId || xid >= snapshot->xmax)
-			{
-
-				/*--------
-				 * Seems that there is no sense to store
-				 * 		xid >= snapshot->xmax
-				 * (what we got from ReadNewTransactionId above)
-				 * in snapshot->xip.  We just assume that all xacts
-				 * with such xid-s are running and may be ignored.
-				 *--------
-				 */
+				!TransactionIdIsNormal(xid) ||
+				TransactionIdFollowsOrEquals(xid, snapshot->xmax))
 				continue;
-			}
-			if (xid < snapshot->xmin)
+
+			if (TransactionIdPrecedes(xid, snapshot->xmin))
 				snapshot->xmin = xid;
 			snapshot->xip[count] = xid;
 			count++;
@@ -338,10 +390,11 @@ GetSnapshotData(bool serializable)
 
 	if (serializable)
 		MyProc->xmin = snapshot->xmin;
-	/* Serializable snapshot must be computed before any other... */
-	Assert(MyProc->xmin != InvalidTransactionId);
 
-	SpinRelease(SInvalLock);
+	LWLockRelease(SInvalLock);
+
+	/* Serializable snapshot must be computed before any other... */
+	Assert(TransactionIdIsValid(MyProc->xmin));
 
 	snapshot->xcnt = count;
 	return snapshot;
@@ -404,7 +457,7 @@ GetUndoRecPtr(void)
 	XLogRecPtr	tempr;
 	int			index;
 
-	SpinAcquire(SInvalLock);
+	LWLockAcquire(SInvalLock, LW_SHARED);
 
 	for (index = 0; index < segP->lastBackend; index++)
 	{
@@ -423,7 +476,35 @@ GetUndoRecPtr(void)
 		}
 	}
 
-	SpinRelease(SInvalLock);
+	LWLockRelease(SInvalLock);
 
 	return (urec);
 }
+
+/*
+ * BackendIdGetProc - given a BackendId, find its PROC structure
+ *
+ * This is a trivial lookup in the ProcState array.  We assume that the caller
+ * knows that the backend isn't going to go away, so we do not bother with
+ * locking.
+ */
+struct PROC *
+BackendIdGetProc(BackendId procId)
+{
+	SISeg	   *segP = shmInvalBuffer;
+
+	if (procId > 0 && procId <= segP->lastBackend)
+	{
+		ProcState  *stateP = &segP->procState[procId - 1];
+		SHMEM_OFFSET pOffset = stateP->procStruct;
+
+		if (pOffset != INVALID_OFFSET)
+		{
+			PROC	   *proc = (PROC *) MAKE_PTR(pOffset);
+
+			return proc;
+		}
+	}
+
+	return NULL;
+}
