@@ -3,7 +3,7 @@
  *
  * Copyright 2000 by PostgreSQL Global Development Group
  *
- * $Header: /home/projects/pgsql/cvsroot/pgsql/src/bin/psql/large_obj.c,v 1.14 2001/03/22 04:00:20 momjian Exp $
+ * $Header: /cvsroot/pgsql-server/src/bin/psql/large_obj.c,v 1.18 2001/10/25 05:49:54 momjian Exp $
  */
 #include "postgres_fe.h"
 #include "large_obj.h"
@@ -16,6 +16,9 @@
 #include "print.h"
 
 
+#define atooid(x)  ((Oid) strtoul((x), NULL, 10))
+
+
 /*
  * Since all large object ops must be in a transaction, we must do some magic
  * here. You can set the variable lo_transaction to one of commit|rollback|
@@ -63,9 +66,9 @@ handle_transaction(void)
 	else if (!QUIET())
 	{
 		if (commit)
-			puts("Warning: Your transaction in progress has been committed.");
+			puts(gettext("Warning: Your transaction in progress has been committed."));
 		else
-			puts("Warning: Your transaction in progress has been rolled back.");
+			puts(gettext("Warning: Your transaction in progress has been rolled back."));
 	}
 
 	PQsetNoticeProcessor(pset.db, old_notice_hook, NULL);
@@ -92,9 +95,7 @@ do_lo_export(const char *loid_arg, const char *filename_arg)
 
 	if (!pset.db)
 	{
-		if (!pset.cur_cmd_interactive)
-			fprintf(stderr, "%s: ", pset.progname);
-		fputs("\\lo_export: not connected to a database\n", stderr);
+		psql_error("\\lo_export: not connected to a database\n");
 		return false;
 	}
 
@@ -109,7 +110,7 @@ do_lo_export(const char *loid_arg, const char *filename_arg)
 		PQclear(res);
 	}
 
-	status = lo_export(pset.db, atol(loid_arg), filename_arg);
+	status = lo_export(pset.db, atooid(loid_arg), filename_arg);
 	if (status != 1)
 	{							/* of course this status is documented
 								 * nowhere :( */
@@ -151,7 +152,7 @@ do_lo_import(const char *filename_arg, const char *comment_arg)
 {
 	PGresult   *res;
 	Oid			loid;
-	char		buf[1024];
+	char		oidbuf[32];
 	unsigned int i;
 	bool		own_transaction = true;
 	const char *var = GetVariable(pset.vars, "LO_TRANSACTION");
@@ -161,9 +162,7 @@ do_lo_import(const char *filename_arg, const char *comment_arg)
 
 	if (!pset.db)
 	{
-		if (!pset.cur_cmd_interactive)
-			fprintf(stderr, "%s: ", pset.progname);
-		fputs("\\lo_import: not connected to a database\n", stderr);
+		psql_error("\\lo_import: not connected to a database\n");
 		return false;
 	}
 
@@ -191,17 +190,16 @@ do_lo_import(const char *filename_arg, const char *comment_arg)
 	}
 
 	/* insert description if given */
-	if (comment_arg)
+	/* XXX don't try to hack pg_description if not superuser */
+	/* XXX ought to replace this with some kind of COMMENT command */
+	if (comment_arg && pset.issuper)
 	{
-		sprintf(buf, "INSERT INTO pg_description VALUES (%u, '", loid);
-		for (i = 0; i < strlen(comment_arg); i++)
-			if (comment_arg[i] == '\'')
-				strcat(buf, "\\'");
-			else
-				strncat(buf, &comment_arg[i], 1);
-		strcat(buf, "')");
+		char	   *cmdbuf;
+		char	   *bufptr;
+		int			slen = strlen(comment_arg);
 
-		if (!(res = PSQLexec(buf)))
+		cmdbuf = malloc(slen * 2 + 256);
+		if (!cmdbuf)
 		{
 			if (own_transaction)
 			{
@@ -210,6 +208,32 @@ do_lo_import(const char *filename_arg, const char *comment_arg)
 			}
 			return false;
 		}
+		sprintf(cmdbuf,
+				"INSERT INTO pg_description VALUES ('%u', "
+		   "(SELECT oid FROM pg_class WHERE relname = 'pg_largeobject'),"
+				" 0, '", loid);
+		bufptr = cmdbuf + strlen(cmdbuf);
+		for (i = 0; i < slen; i++)
+		{
+			if (comment_arg[i] == '\'' || comment_arg[i] == '\\')
+				*bufptr++ = '\\';
+			*bufptr++ = comment_arg[i];
+		}
+		strcpy(bufptr, "')");
+
+		if (!(res = PSQLexec(cmdbuf)))
+		{
+			if (own_transaction)
+			{
+				res = PQexec(pset.db, "ROLLBACK");
+				PQclear(res);
+			}
+			free(cmdbuf);
+			return false;
+		}
+
+		PQclear(res);
+		free(cmdbuf);
 	}
 
 	if (own_transaction)
@@ -225,9 +249,9 @@ do_lo_import(const char *filename_arg, const char *comment_arg)
 	}
 
 
-	fprintf(pset.queryFout, "lo_import %d\n", loid);
-	sprintf(buf, "%u", (unsigned int) loid);
-	SetVariable(pset.vars, "LASTOID", buf);
+	fprintf(pset.queryFout, "lo_import %u\n", loid);
+	sprintf(oidbuf, "%u", loid);
+	SetVariable(pset.vars, "LASTOID", oidbuf);
 
 	return true;
 }
@@ -244,7 +268,7 @@ do_lo_unlink(const char *loid_arg)
 {
 	PGresult   *res;
 	int			status;
-	Oid			loid = (Oid) atol(loid_arg);
+	Oid			loid = atooid(loid_arg);
 	char		buf[256];
 	bool		own_transaction = true;
 	const char *var = GetVariable(pset.vars, "LO_TRANSACTION");
@@ -254,9 +278,7 @@ do_lo_unlink(const char *loid_arg)
 
 	if (!pset.db)
 	{
-		if (!pset.cur_cmd_interactive)
-			fprintf(stderr, "%s: ", pset.progname);
-		fputs("\\lo_unlink: not connected to a database\n", stderr);
+		psql_error("\\lo_unlink: not connected to a database\n");
 		return false;
 	}
 
@@ -284,18 +306,24 @@ do_lo_unlink(const char *loid_arg)
 	}
 
 	/* remove the comment as well */
-	sprintf(buf, "DELETE FROM pg_description WHERE objoid = %u", loid);
-	if (!(res = PSQLexec(buf)))
+	/* XXX don't try to hack pg_description if not superuser */
+	/* XXX ought to replace this with some kind of COMMENT command */
+	if (pset.issuper)
 	{
-		if (own_transaction)
+		sprintf(buf, "DELETE FROM pg_description WHERE objoid = '%u' "
+				"AND classoid = (SELECT oid FROM pg_class WHERE relname = 'pg_largeobject')",
+				loid);
+		if (!(res = PSQLexec(buf)))
 		{
-			res = PQexec(pset.db, "ROLLBACK");
-			PQclear(res);
+			if (own_transaction)
+			{
+				res = PQexec(pset.db, "ROLLBACK");
+				PQclear(res);
+			}
+			return false;
 		}
-		return false;
 	}
 
-
 	if (own_transaction)
 	{
 		if (!(res = PSQLexec("COMMIT")))
@@ -308,7 +336,7 @@ do_lo_unlink(const char *loid_arg)
 	}
 
 
-	fprintf(pset.queryFout, "lo_unlink %d\n", loid);
+	fprintf(pset.queryFout, "lo_unlink %u\n", loid);
 
 	return true;
 }
@@ -327,10 +355,11 @@ do_lo_list(void)
 	char		buf[1024];
 	printQueryOpt myopt = pset.popt;
 
-	strcpy(buf,
-	  "SELECT loid as \"ID\", obj_description(loid) as \"Description\"\n"
-		   "FROM (SELECT DISTINCT loid FROM pg_largeobject) x\n"
-		   "ORDER BY \"ID\"");
+	snprintf(buf, sizeof(buf),
+			 "SELECT loid as \"ID\", obj_description(loid, 'pg_largeobject') as \"%s\"\n"
+			 "FROM (SELECT DISTINCT loid FROM pg_largeobject) x\n"
+			 "ORDER BY \"ID\"",
+			 gettext("Description"));
 
 	res = PSQLexec(buf);
 	if (!res)
@@ -338,7 +367,7 @@ do_lo_list(void)
 
 	myopt.topt.tuples_only = false;
 	myopt.nullPrint = NULL;
-	myopt.title = "Large objects";
+	myopt.title = gettext("Large objects");
 
 	printQuery(res, &myopt, pset.queryFout);
 
