@@ -3,7 +3,7 @@
  *			  procedural language
  *
  * IDENTIFICATION
- *	  $Header: /usr/local/cvsroot/pgsql/src/pl/plpgsql/src/pl_handler.c,v 1.2 1998/09/01 04:40:25 momjian Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/pl/plpgsql/src/pl_handler.c,v 1.10 2001/10/25 05:50:20 momjian Exp $
  *
  *	  This software is copyrighted by Jan Wieck - Hamburg.
  *
@@ -45,65 +45,102 @@
 #include "plpgsql.h"
 #include "pl.tab.h"
 
-#include "executor/spi.h"
-#include "commands/trigger.h"
-#include "utils/elog.h"
-#include "utils/builtins.h"
-#include "fmgr.h"
 #include "access/heapam.h"
-
-#include "utils/syscache.h"
 #include "catalog/pg_proc.h"
 #include "catalog/pg_type.h"
+#include "utils/builtins.h"
+#include "utils/syscache.h"
 
 
+/*
+ * Head of list of already-compiled functions
+ */
 static PLpgSQL_function *compiled_functions = NULL;
 
 
-Datum plpgsql_call_handler(FmgrInfo *proinfo,
-					 FmgrValues *proargs, bool *isNull);
-
-static Datum plpgsql_func_handler(FmgrInfo *proinfo,
-					 FmgrValues *proargs, bool *isNull);
-
-static HeapTuple plpgsql_trigger_handler(FmgrInfo *proinfo);
+static bool func_up_to_date(PLpgSQL_function * func);
 
 
 /* ----------
- * plpgsql_call_handler		- This is the only visible function
- *				  of the PL interpreter. The PostgreSQL
- *				  function manager and trigger manager
- *				  call this function for execution of
- *				  PL/pgSQL procedures.
+ * plpgsql_call_handler
+ *
+ * This is the only visible function of the PL interpreter.
+ * The PostgreSQL function manager and trigger manager
+ * call this function for execution of PL/pgSQL procedures.
  * ----------
  */
+PG_FUNCTION_INFO_V1(plpgsql_call_handler);
+
 Datum
-plpgsql_call_handler(FmgrInfo *proinfo,
-					 FmgrValues *proargs,
-					 bool *isNull)
+plpgsql_call_handler(PG_FUNCTION_ARGS)
 {
+	bool		isTrigger = CALLED_AS_TRIGGER(fcinfo);
+	Oid			funcOid = fcinfo->flinfo->fn_oid;
+	PLpgSQL_function *func;
 	Datum		retval;
 
-	/* ----------
+	/*
 	 * Connect to SPI manager
-	 * ----------
 	 */
 	if (SPI_connect() != SPI_OK_CONNECT)
 		elog(ERROR, "plpgsql: cannot connect to SPI manager");
 
-	/* ----------
-	 * Determine if called as function or trigger and
-	 * call appropriate subhandler
-	 * ----------
+	/*
+	 * Check if we already compiled this function and saved the pointer
+	 * (ie, current FmgrInfo has been used before)
 	 */
-	if (CurrentTriggerData == NULL)
-		retval = plpgsql_func_handler(proinfo, proargs, isNull);
-	else
-		retval = (Datum) plpgsql_trigger_handler(proinfo);
+	func = (PLpgSQL_function *) fcinfo->flinfo->fn_extra;
+	if (func != NULL)
+	{
+		Assert(func->fn_oid == funcOid);
 
-	/* ----------
+		/*
+		 * But is the function still up to date?
+		 */
+		if (!func_up_to_date(func))
+			func = NULL;
+	}
+
+	if (func == NULL)
+	{
+		/*
+		 * Check if we already compiled this function for another caller
+		 */
+		for (func = compiled_functions; func != NULL; func = func->next)
+		{
+			if (funcOid == func->fn_oid && func_up_to_date(func))
+				break;
+		}
+
+		/*
+		 * If not, do so and add it to the compiled ones
+		 */
+		if (func == NULL)
+		{
+			func = plpgsql_compile(funcOid,
+								   isTrigger ? T_TRIGGER : T_FUNCTION);
+			func->next = compiled_functions;
+			compiled_functions = func;
+		}
+
+		/*
+		 * Save pointer in FmgrInfo to avoid search on subsequent calls
+		 */
+		fcinfo->flinfo->fn_extra = (void *) func;
+	}
+
+	/*
+	 * Determine if called as function or trigger and call appropriate
+	 * subhandler
+	 */
+	if (isTrigger)
+		retval = PointerGetDatum(plpgsql_exec_trigger(func,
+									   (TriggerData *) fcinfo->context));
+	else
+		retval = plpgsql_exec_function(func, fcinfo);
+
+	/*
 	 * Disconnect from SPI manager
-	 * ----------
 	 */
 	if (SPI_finish() != SPI_OK_FINISH)
 		elog(ERROR, "plpgsql: SPI_finish() failed");
@@ -112,81 +149,28 @@ plpgsql_call_handler(FmgrInfo *proinfo,
 }
 
 
-/* ----------
- * plpgsql_func_handler()	- Handler for regular function calls
- * ----------
+/*
+ * Check to see if a compiled function is still up-to-date.  This
+ * is needed because CREATE OR REPLACE FUNCTION can modify the
+ * function's pg_proc entry without changing its OID.
  */
-static Datum
-plpgsql_func_handler(FmgrInfo *proinfo,
-					 FmgrValues *proargs,
-					 bool *isNull)
+static bool
+func_up_to_date(PLpgSQL_function * func)
 {
-	PLpgSQL_function *func;
+	HeapTuple	procTup;
+	bool		result;
 
-	/* ----------
-	 * Check if we already compiled this function
-	 * ----------
-	 */
-	for (func = compiled_functions; func != NULL; func = func->next)
-	{
-		if (proinfo->fn_oid == func->fn_oid)
-			break;
-	}
+	procTup = SearchSysCache(PROCOID,
+							 ObjectIdGetDatum(func->fn_oid),
+							 0, 0, 0);
+	if (!HeapTupleIsValid(procTup))
+		elog(ERROR, "plpgsql: cache lookup for proc %u failed",
+			 func->fn_oid);
 
-	/* ----------
-	 * If not, do so and add it to the compiled ones
-	 * ----------
-	 */
-	if (func == NULL)
-	{
-		func = plpgsql_compile(proinfo->fn_oid, T_FUNCTION);
+	result = (func->fn_xmin == procTup->t_data->t_xmin &&
+			  func->fn_cmin == procTup->t_data->t_cmin);
 
-		func->next = compiled_functions;
-		compiled_functions = func;
-	}
+	ReleaseSysCache(procTup);
 
-	return plpgsql_exec_function(func, proargs, isNull);
-}
-
-
-/* ----------
- * plpgsql_trigger_handler()	- Handler for trigger calls
- * ----------
- */
-static HeapTuple
-plpgsql_trigger_handler(FmgrInfo *proinfo)
-{
-	TriggerData *trigdata;
-	PLpgSQL_function *func;
-
-	/* ----------
-	 * Save the current trigger data local
-	 * ----------
-	 */
-	trigdata = CurrentTriggerData;
-	CurrentTriggerData = NULL;
-
-	/* ----------
-	 * Check if we already compiled this trigger procedure
-	 * ----------
-	 */
-	for (func = compiled_functions; func != NULL; func = func->next)
-	{
-		if (proinfo->fn_oid == func->fn_oid)
-			break;
-	}
-
-	/* ----------
-	 * If not, do so and add it to the compiled ones
-	 * ----------
-	 */
-	if (func == NULL)
-	{
-		func = plpgsql_compile(proinfo->fn_oid, T_TRIGGER);
-
-		func->next = compiled_functions;
-		compiled_functions = func;
-	}
-
-	return plpgsql_exec_trigger(func, trigdata);
+	return result;
 }
